import numpy as np

from rock_physics_open.equinor_utilities.various_utilities.types import (
    Array1D,
    Array2D,
    Array3D,
)


def calc_theta_vec(
    v: Array2D[np.float64],
    omega: float,
    gamma: Array2D[np.float64],
    tau: Array1D[np.float64],
    kd_uuvv: Array2D[np.float64],
    dr: Array1D[np.float64],
    k: Array1D[np.float64],
    kappa: Array1D[np.float64],
    kappa_f: Array1D[np.float64],
) -> Array3D[np.complex128]:
    """Returns the theta tensor (6x6 matrix) for more explanation see e.g.
    Agersborg et al. 2009 or "The effects of drained and undrained loading in
    visco-elastic waves in rock-like composites" M. Jakobsen and T.A. Johansen.
    (2005). Int. J. Solids and Structures (42). p. 1597-1611.

    Parameters
    ----------
    v  : np.ndarray
        Concentration of all the empty cavities (1x(numbers of empty cavities) vector).
    omega : np.ndarray
        Frequency (2*pi*f).
    gamma : np.ndarray
        Gamma factor of all the inclusions (1x(numbers of empty cavities) vector).
    tau : np.ndarray
        Relaxation time constant (1x(numbers of empty cavities) vector).
    kd_uuvv : np.ndarray
        Tensor sum of the dry K tensor for all the cavities (1x(numbers of empty cavities) vector).
    dr : np.ndarray
        Permeability/viscosity.
    k : np.ndarray
        Wave number vector.
    kappa : np.ndarray
        Bulk modulus of the host material.
    kappa_f : np.ndarray
        Bulk modulus of the fluid.

    Returns
    -------
    np.ndarray
        Theta tensor.
    """
    sigma_a = np.sum((v / (1 + 1j * omega * gamma * tau)), axis=1)
    sigma_b = np.sum((v / (1 + 1j * omega * gamma * tau)) * kd_uuvv, axis=1)

    return kappa_f / (
        (1 - kappa_f / kappa) * sigma_a
        + kappa_f * sigma_b
        - (1j * k * k / omega) * dr * kappa_f
    )



import numpy as np

from rock_physics_open.equinor_utilities.snapshot_test_utilities import (
    get_snapshot_name,
)
from rock_physics_open.equinor_utilities.various_utilities import vp_vs_rho_stats


def test_vp_vs_rho_stats():
    rng = np.random.default_rng(123456)
    vp_obs = rng.uniform(low=0.0, high=1.0, size=100)
    vs_obs = rng.uniform(low=0.0, high=1.0, size=100)
    rho_obs = rng.uniform(low=0.0, high=1.0, size=100)
    vp_est = rng.uniform(low=0.0, high=1.0, size=100)
    vs_est = rng.uniform(low=0.0, high=1.0, size=100)
    rho_est = rng.uniform(low=0.0, high=1.0, size=100)
    # fname = os.path.join(os.path.dirname(__file__), "snapshots", "vp_vs_rho_stats.csv")
    fname = get_snapshot_name(include_extension=False, include_filename=False) + ".csv"
    file_open_mode = "w"
    disp_res = False
    vp_vs_rho_stats(
        vp_obs,
        vs_obs,
        rho_obs,
        vp_est,
        vs_est,
        rho_est,
        fname,
        "pytest",
        "pytest_well",
        file_mode=file_open_mode,
        disp_results=disp_res,
    )


def test_multi_vp_vs_rho_stats():
    rng = np.random.default_rng(123456)
    vp_obs = [rng.uniform(low=0.0, high=1.0, size=100)] * 3
    vs_obs = [rng.uniform(low=0.0, high=1.0, size=100)] * 3
    rho_obs = [rng.uniform(low=0.0, high=1.0, size=100)] * 3
    vp_est = [rng.uniform(low=0.0, high=1.0, size=100)] * 3
    vs_est = [rng.uniform(low=0.0, high=1.0, size=100)] * 3
    rho_est = [rng.uniform(low=0.0, high=1.0, size=100)] * 3
    set_names = ["pytest_1", "pytest_2", "pytest_3"]
    well_names = ["pytest_well_1", "pytest_well_2", "pytest_well_3"]
    fname = get_snapshot_name(include_extension=False, include_filename=False) + ".csv"
    file_open_mode = "w"
    disp_res = False
    vp_vs_rho_stats(
        vp_obs,
        vs_obs,
        rho_obs,
        vp_est,
        vs_est,
        rho_est,
        fname,
        set_names,
        well_names,
        file_mode=file_open_mode,
        disp_results=disp_res,
    )



import numpy as np

from rock_physics_open.equinor_utilities.various_utilities.types import (
    Array1D,
    Array2D,
    Array3D,
    Array4D,
)

from .array_functions import array_matrix_mult
from .check_and_tile import check_and_tile


def calc_z_vec(
    s0: Array3D[np.float64],
    td: Array4D[np.float64],
    td_bar: Array4D[np.float64],
    omega: float,
    gamma: Array2D[np.float64],
    v: Array2D[np.float64],
    tau: Array1D[np.float64],
) -> tuple[Array4D[np.complex128], Array4D[np.complex128]]:
    """Returns the z tensor (6x6x(numbers of empty cavities) matrix) for more explanation see e.g.
    Agersborg et al. 2009 or "The effects of drained and undrained loading in
    visco-elastic waves in rock-like composites" M. Jakobsen and T.A. Johansen.
    (2005). Int. J. Solids and Structures (42). p. 1597-1611.

    Parameters
    ----------
    s0  : np.ndarray
        Stiffness tensor of the host material (6x6 matrix).
    td  : np.ndarray
        Dry t-matrix tensors, (6x6x(numbers of empty cavities) matrix).
    td_bar  : np.ndarray
        Dry t-matrix tensors, (6x6x(numbers of empty cavities) matrix).
    omega : np.ndarray
        Frequency (2*pi*f).
    gamma : np.ndarray
        Gamma factor of all the inclusions (1x(numbers of empty cavities) vector).
    v  :  np.ndarray
        Concentration of all the empty cavities (1x(numbers of empty cavities) vector).
    tau  :  np.ndarray
        Relaxation time constant (1x(numbers of empty cavities) vector).

    Returns
    -------
    tuple
        z, z_bar : (np.ndarray, np.ndarray).
    """
    i2_i2, log_length, alpha_length = check_and_tile(s0, td)

    sum_z = 0.0
    z = np.zeros((log_length, 6, 6, alpha_length), dtype="complex128")
    z_bar = np.zeros((log_length, 6, 6, alpha_length), dtype="complex128")

    for j in range(alpha_length):
        sum_z = sum_z + (td[:, :, :, j] + td_bar[:, :, :, j]) * (
            v[:, j] / (1 + 1j * omega * gamma[:, j] * tau[j])
        ).reshape(log_length, 1, 1)

    for j in range(alpha_length):
        z[:, :, :, j] = array_matrix_mult(
            td[:, :, :, j],
            s0,
            i2_i2,
            s0,
            sum_z,  # pyright: ignore[reportArgumentType] | sum_z is 3D array after summation
        )
        z_bar[:, :, :, j] = array_matrix_mult(
            td_bar[:, :, :, j],
            s0,
            i2_i2,
            s0,
            sum_z,  # pyright: ignore[reportArgumentType] | sum_z is 3D array after summation
        )

    return z, z_bar



import numpy as np
import numpy.typing as npt


def posterior_probability(
    min_dist: npt.NDArray[np.float64], dist: npt.NDArray[np.float64]
) -> npt.NDArray[np.float64]:
    """
    Posterior probability, which is defined as the exponential of minimum distance divided by the sum of the
    exponentials of distance to all classes.

    Parameters
    ----------
    min_dist : np.ndarray
        Minimum class distance according to some metric.
    dist : np.ndarray
        All class distances, each class in a column in a two-dimensional array.

    Returns
    -------
    np.ndarray
        Posterior probability array.
    """
    dist *= -1.0
    n_exp = np.exp(dist)
    d_sum = n_exp.sum(axis=1)
    return np.exp(-min_dist) / d_sum



import numpy as np
from numpy.random import default_rng

from rock_physics_open.equinor_utilities.std_functions import aki_richards, smith_gidlow

rg = default_rng(12345)
vp = 3500 * (1.0 + 0.2 * rg.random(11))
vs = 1200 * (1.0 + 0.4 * rg.random(11))
rho = 2650 * (1.0 + 0.02 * rg.random(11))
theta = np.ones(11)
k = 1.9


class TestReflectivity:
    def test_aki_richards(self):
        r_aki_rich = aki_richards(vp, vs, rho, theta, k=2.0)
        r_aki_rich_ref = np.array(
            [
                0.0048749,
                0.0206719,
                -0.0025819,
                -0.015988,
                -0.0030281,
                0.0119779,
                -0.0165907,
                0.023673,
                0.0103467,
                -0.0293775,
                -0.0293775,
            ]
        )
        np.testing.assert_almost_equal(r_aki_rich, r_aki_rich_ref)

    def test_smith_gidlow(self):
        r_sm_gi = smith_gidlow(vp, vs, rho, theta, k=2.0)
        r_sm_gi_ref = np.array(
            [
                0.0053144,
                0.027062,
                -0.0066175,
                -0.0161272,
                -0.0033894,
                0.0152015,
                -0.0238758,
                0.0280036,
                0.0144885,
                -0.038749,
                -0.038749,
            ]
        )
        np.testing.assert_almost_equal(r_sm_gi, r_sm_gi_ref)



from .class_stats import gen_class_stats
from .lin_class import lin_class
from .mahal_class import mahal_class
from .norm_class import norm_class
from .poly_class import poly_class
from .post_prob import posterior_probability
from .two_step_classification import gen_two_step_class_stats

__all__ = [
    "gen_class_stats",
    "lin_class",
    "mahal_class",
    "norm_class",
    "poly_class",
    "posterior_probability",
    "gen_two_step_class_stats",
]



import numpy as np
import numpy.typing as npt


def bp_standing(
    density: npt.NDArray[np.float64],
    gas_oil_ratio: npt.NDArray[np.float64],
    gas_gravity: npt.NDArray[np.float64],
    temperature: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    Reservoir oils include some natural gas in solution. The pressure at which
    this natural gas begins to come out of solution and form bubbles is known
    as the bubble point pressure. See https://petrowiki.org/Oil_bubblepoint_pressure
    Based on the correlation from:
    Standing, M. B. "A pressure-volume-temperature correlation for mixtures of
    California oils and gases." Drilling and Production Practice. American
    Petroleum Institute, 1947.
    Uses refinement described here: https://petrowiki.org/Oil_bubblepoint_pressure
    :param density: density of oil at room conditions [kg/m^3]
    :param gas_oil_ratio: The volume ratio of gas to oil [l/l]
    :param gas_gravity: molar mass of gas relative to air molar mas.
    :param temperature: temperature of oil [°C]
    :return: bubble point pressure [Pa]
    """

    # Standing, M.B. (1947) uses:
    # * pressure in psi
    # * temperature in F
    # * gas oil ratio in cu ft per bbl
    # * bubble point in in absolute psi
    # * gravity_of_tank_oil in API
    #
    # The paper describes that bubble point is a function of
    # (gor / gr) ** 0.83
    #   * (10**(0.00091 * t)/10**(0.0125*gravity_of_tank_oil))
    #
    # and gives the following example:
    #   Bubble point pressure at 200° F
    #   of a liquid having gas-oil ratio
    #   350 CFB, a gas gravity 0.75, and
    #   a tank oil gravity of 30° API.
    #   The required pressure is found
    #   to be 1930 psia.
    #
    # We could scale to fit this example, however,
    # we use 1896 psia.
    #
    # For density in kg/m^3:
    # 10**(0.0125*gravity_of_tank_oil)=
    #        10**(1770.79/density - 1.64375)=
    #         e**(4072.69738323/density - 3.78487)
    # For temperature in Celsius:
    #    10**(0.00091 * Fahrenheit) =
    #    10**(0.001638*Celsius + 0.02912) =
    #     e**(0.00377163*Celsius + 0.0670513)
    #
    # Removing constants factors we get
    # (gor_ratio /
    #  gas_gravity)) ** 0.83 / ( np.exp(4072.69738323 / density -
    #  0.00377163438 * temperature_c)
    #
    # The equation occurs in this form as equation 21a in
    # Batzle, Michael, and Zhijing Wang. "Seismic properties of pore fluids."
    # Geophysics 57.11 (1992): 1396-1408.

    ratio = gas_oil_ratio / gas_gravity
    denominator = np.exp(4072 / density - 0.00377 * temperature)

    return 24469793.9134 * ratio**0.83 / denominator



from typing import overload

import numpy as np
import numpy.typing as npt

from rock_physics_open.equinor_utilities.various_utilities.types import (
    Array1D,
    Array2D,
    TMatrixCallable,
)

from .parse_t_matrix_inputs import parse_t_matrix_inputs
from .t_matrix_vector import calc_pressure_vec, pressure_input_utility


@overload
def run_t_matrix(
    k_min: npt.NDArray[np.float64],
    mu_min: npt.NDArray[np.float64],
    rho_min: npt.NDArray[np.float64],
    k_fl: npt.NDArray[np.float64],
    rho_fl: npt.NDArray[np.float64],
    phi: npt.NDArray[np.float64],
    perm: npt.NDArray[np.float64] | float,
    visco: npt.NDArray[np.float64] | float,
    alpha: npt.NDArray[np.float64],
    v: npt.NDArray[np.float64],
    tau: npt.NDArray[np.float64],
    frequency: float,
    angle: float,
    frac_inc_con: npt.NDArray[np.float64] | float,
    frac_inc_ani: npt.NDArray[np.float64] | float,
    pressure: npt.NDArray[np.float64],
    scenario: int | None = None,
    fcn: TMatrixCallable | str | None = None,
) -> list[Array2D[np.float64]]: ...


@overload
def run_t_matrix(
    k_min: npt.NDArray[np.float64],
    mu_min: npt.NDArray[np.float64],
    rho_min: npt.NDArray[np.float64],
    k_fl: npt.NDArray[np.float64],
    rho_fl: npt.NDArray[np.float64],
    phi: npt.NDArray[np.float64],
    perm: npt.NDArray[np.float64] | float,
    visco: npt.NDArray[np.float64] | float,
    alpha: npt.NDArray[np.float64],
    v: npt.NDArray[np.float64],
    tau: npt.NDArray[np.float64],
    frequency: float,
    angle: float,
    frac_inc_con: npt.NDArray[np.float64] | float,
    frac_inc_ani: npt.NDArray[np.float64] | float,
    pressure: None = None,
    scenario: int | None = None,
    fcn: TMatrixCallable | str | None = None,
) -> tuple[
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
]: ...


def run_t_matrix(
    k_min: npt.NDArray[np.float64],
    mu_min: npt.NDArray[np.float64],
    rho_min: npt.NDArray[np.float64],
    k_fl: npt.NDArray[np.float64],
    rho_fl: npt.NDArray[np.float64],
    phi: npt.NDArray[np.float64],
    perm: npt.NDArray[np.float64] | float,
    visco: npt.NDArray[np.float64] | float,
    alpha: npt.NDArray[np.float64],
    v: npt.NDArray[np.float64],
    tau: npt.NDArray[np.float64],
    frequency: float,
    angle: float,
    frac_inc_con: npt.NDArray[np.float64] | float,
    frac_inc_ani: npt.NDArray[np.float64] | float,
    pressure: npt.NDArray[np.float64] | None = None,
    scenario: int | None = None,
    fcn: TMatrixCallable | str | None = None,
) -> (
    list[Array2D[np.float64]]
    | tuple[
        Array1D[np.float64],
        Array1D[np.float64],
        Array1D[np.float64],
        Array1D[np.float64],
    ]
):
    """Function to run T-Matrix in different flavours, with or without pressure steps included.
    A frontend to running T-Matrix model, including testing of all input parameters in the parse_t_matrix_inputs.

    Parameters
    ----------
    k_min : np.ndarray
        N length numpy array, mineral bulk modulus [Pa].
    mu_min : np.ndarray
        N length numpy array, mineral shear modulus [Pa].
    rho_min : np.ndarray
        N length numpy array, mineral density [kg/m^3].
    k_fl : np.ndarray
        N length numpy array, fluid bulk modulus [Pa].
    rho_fl : np.ndarray
        N length numpy array, mineral density [kg/m^3].
    phi : np.ndarray or float
        N length numpy array, total porosity [ratio].
    perm : np.ndarray or float
        Float or N length numpy array, permeability [mD].
    visco : np.ndarray or float
        Float or N length numpy array, fluid viscosity [cP].
    alpha : np.ndarray
        M or NxM length numpy array, inclusion aspect ratio [ratio].
    v : np.ndarray
        M or NxM length numpy array, inclusion concentration [ratio].
    tau : np.ndarray
        M length numpy array, relaxation time [s].
    frequency : float
        Single float, signal frequency [Hz].
    angle : float
        Single float, angle of symmetry plane [degree].
    frac_inc_con : np.ndarray or float
        Single float or N length numpy array, fraction of connected inclusions [ratio].
    frac_inc_ani : np.ndarray or float
        Single float or N length numpy array, fraction of anisotropic inclusions [ratio].
    pressure : [type], optional
        > 1 value list or numpy array in ascending order, effective pressure [Pa], by default None.
    scenario : int, optional
        Pre-set scenarios for alpha, v and tau, by default None.
    fcn : callable | str, optional
        Function with which to run the T-Matrix model or string with function name within t_matrix_models, by default None.

    Returns
    -------
    tuple
        vp, vsv, vsh, rho: (np.ndarray, np.ndarray, np.ndarray, np.ndarray).
    """
    # Check all input parameters and make sure that they are on expected format and shape
    (
        k_min,
        mu_min,
        rho_min,
        k_fl,
        rho_fl,
        phi_,
        perm_,
        visco_,
        alpha,
        v,
        tau,
        frequency,
        angle,
        frac_inc_con_,
        frac_inc_ani_,
        pressure,
        fcn,
        ctrl_connected,
        _,
    ) = parse_t_matrix_inputs(
        k_min=k_min,
        mu_min=mu_min,
        rho_min=rho_min,
        k_fl=k_fl,
        rho_fl=rho_fl,
        phi=phi,
        perm=perm,
        visco=visco,
        alpha=alpha,
        v=v,
        tau=tau,
        frequency=frequency,
        angle=angle,
        frac_inc_con=frac_inc_con,
        frac_inc_ani=frac_inc_ani,
        pressure=pressure,
        scenario=scenario,
        fcn=fcn,
    )

    # If there are no pressure steps to consider the whole task can be assigned to the T-Matrix function
    if pressure is None:
        vp, vsv, vsh, rho = fcn(
            k_min=k_min,
            mu_min=mu_min,
            rho_min=rho_min,
            k_fl=k_fl,
            rho_fl=rho_fl,
            phi=phi_,
            perm=perm_,
            visco=visco_,
            alpha=alpha,
            v=v,
            tau=tau,
            frequency=frequency,
            angle=angle,
            frac_inc_con=frac_inc_con_,
            frac_inc_ani=frac_inc_ani_,
        )
        return vp, vsv, vsh, rho

    # Pressure steps is mimicking depletion, and inclusion shape parameters can be changed
    # In Remy's implementation it is generally possible to have different alpha's and v's for connected and
    # isolated part, but this is not included directly in Calculo's implementation in C++. For the time being,
    # this possibility is not included here

    log_length = phi_.shape[0]
    pressure_steps = pressure.shape[0]
    # It is the change from initial pressure (first value in the pressure vector) that is input to the function
    # that estimates matrix pressure sensitivity
    delta_pres = np.diff(pressure)

    # Predefine output vectors
    vp = np.zeros((log_length, pressure_steps))
    vs_v = np.zeros((log_length, pressure_steps))
    vs_h = np.zeros((log_length, pressure_steps))
    rho_b_est = np.zeros((log_length, pressure_steps))

    # Matrix properties needed as input to calc_pressure_vec
    c0, s0, gd = pressure_input_utility(k_min, mu_min, log_length)

    for i in range(pressure_steps):
        vp[:, i], vs_v[:, i], vs_h[:, i], rho_b_est[:, i] = fcn(
            k_min=k_min,
            mu_min=mu_min,
            rho_min=rho_min,
            k_fl=k_fl,
            rho_fl=rho_fl,
            phi=phi_,
            perm=perm_,
            visco=visco_,
            alpha=alpha,
            v=v,
            tau=tau,
            frequency=frequency,
            angle=angle,
            frac_inc_con=frac_inc_con_,
            frac_inc_ani=frac_inc_ani_,
        )

        if i != pressure_steps - 1:
            # Need to check for extreme cases with all connected or all isolated
            if ctrl_connected != 0:
                v_con = v * (phi * frac_inc_con).reshape(log_length, 1)
            else:
                v_con = np.zeros_like(v)
            if ctrl_connected != 2:
                v_iso = v * (phi * (1 - frac_inc_con)).reshape(log_length, 1)
            else:
                v_iso = np.zeros_like(v)
            alpha_con, v_con, alpha_iso, v_iso, tau, _ = calc_pressure_vec(
                alpha_con=alpha,
                alpha_iso=alpha,
                v_con=v_con,
                v_iso=v_iso,
                c0=c0,
                s_0=s0,
                gd=gd,
                d_p=delta_pres[i],
                tau=tau,
                gamma=np.zeros_like(tau),
                k_fl=k_fl,
                ctrl=ctrl_connected,
                frac_ani=float(frac_inc_ani),
            )

            # Post-process outputs from calc_pressure_vec to match required inputs to the T-Matrix function
            # frac_inc_con, v and alpha are inputs that need to be updated
            # As mentioned: assume that there is no distinction between alpha_con and alpha_iso, that could only
            # happen if they differed from the start

            # v_con, alpha_con or v_iso, alpha_iso can be returned as None from calc_pressure_vec
            if ctrl_connected == 0:
                v_con = np.zeros_like(v_iso)
                alpha = alpha_iso
            elif ctrl_connected == 2:
                alpha = alpha_con
                v_iso = np.zeros_like(v_con)
            else:
                alpha = alpha_con
            # Don't divide by zero
            idx_zero = phi == 0
            if np.any(idx_zero):
                v = v_con + v_iso
                no_zero = np.sum(idx_zero)
                v[~idx_zero, :] = v[~idx_zero, :] / phi[~idx_zero].reshape(
                    log_length - no_zero, 1
                )
            else:
                v = (v_con + v_iso) / phi.reshape(log_length, 1)

            # frac_inc_con is close to a single value, only numerical difference in the range 10^-16, can safely
            # take the average over the alphas to make it match expected input shape
            frac_inc_con = np.mean(v_con / (v_con + v_iso), axis=1)

    # Return variables for each pressure step
    vp_out: list[Array2D[np.float64]] = []
    vs_v_out: list[Array2D[np.float64]] = []
    vs_h_out: list[Array2D[np.float64]] = []

    # Only one rho is needed - no change
    rho_out = rho_b_est[:, 0]

    for i in range(pressure_steps):
        vp_out.append(vp[:, i])
        vs_v_out.append(vs_v[:, i])
        vs_h_out.append(vs_h[:, i])
    return vp_out + vs_v_out + vs_h_out + [rho_out]



from typing import cast

import numpy as np
import numpy.typing as npt

from rock_physics_open.equinor_utilities import gen_utilities, std_functions
from rock_physics_open.sandstone_models.friable_models import (
    CoordinateNumberFunction,
    friable_model,
)


def friable_shaly_sand_sandy_shale_model(
    k_sst: npt.NDArray[np.float64],
    mu_sst: npt.NDArray[np.float64],
    rho_sst: npt.NDArray[np.float64],
    k_mud: npt.NDArray[np.float64],
    mu_mud: npt.NDArray[np.float64],
    rho_mud: npt.NDArray[np.float64],
    k_fl_sst: npt.NDArray[np.float64],
    rho_fl_sst: npt.NDArray[np.float64],
    k_fl_mud: npt.NDArray[np.float64],
    rho_fl_mud: npt.NDArray[np.float64],
    phi: npt.NDArray[np.float64],
    p_eff_sst: npt.NDArray[np.float64],
    p_eff_mud: npt.NDArray[np.float64],
    shale_frac: npt.NDArray[np.float64],
    phi_c_sst: float,
    phi_c_mud: float,
    phi_intr_mud: float,
    coord_num_func_sst: CoordinateNumberFunction,
    n_sst: float,
    coord_num_func_mud: CoordinateNumberFunction,
    n_mud: float,
    shear_red_sst: float,
    shear_red_mud: float,
) -> tuple[
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
]:
    """
    Model for mixing of friable sand and friable shale.

    It is no point to use this model to calculate the shale response only,
    in that case Friable Model with shale parameters does the job.

    The shale fluid should be brine.

    Shale fraction shaleFrac is in the range 0 to 1. For shaleFrac = 0 we
    have a pure sand end member with phi = phiC for sand. For shaleFrac = 1
    we have pure shale with phi = intrinsic porosity. For shaleFrac < phiC
    the model is on the shaly sand trend, for shaleFrac > phiC it is on the
    sandy shale trend.

    Parameters
    ----------
    k_sst : np.ndarray
        Sandstone bulk modulus [Pa].
    mu_sst : np.ndarray
        Sandstone shear modulus [Pa].
    rho_sst : np.ndarray
        Sandstone bulk density [kg/m^3].
    k_mud : np.ndarray
        Shale bulk modulus [Pa].
    mu_mud : np.ndarray
        Shale shear modulus [Pa].
    rho_mud : np.ndarray
        Shale bulk density [kg/m^3].
    k_fl_sst : np.ndarray
        Fluid bulk modulus for sandstone fluid [Pa].
    rho_fl_sst : np.ndarray
        Fluid bulk density for sandstone fluid [kg/m^3].
    k_fl_mud : np.ndarray
        Fluid bulk modulus for shale fluid [Pa].
    rho_fl_mud : np.ndarray
        Fluid bulk density for shale fluid[kg/m^3].
    phi : np.ndarray
        Total porosity [fraction].
    p_eff_sst : np.ndarray
        Effective pressure in sandstone [Pa].
    p_eff_mud : np.ndarray
        Effective pressure in mud [Pa].
    shale_frac : np.ndarray
        Shale fraction [fraction].
    phi_c_sst : float
        Critical porosity for sandstone [fraction].
    phi_c_mud : float
        Critical porosity for mud [fraction].
    phi_intr_mud : float
        Intrinsic shale porosity [fraction].
    coord_num_func_sst : str
        Indication if coordination number should be calculated from porosity or kept constant for sandstone.
    coord_num_func_mud : str
        Indication if coordination number should be calculated from porosity or kept constant for shale.
    n_sst : float
        Coordination number for sandstone [unitless].
    n_mud : float
        Coordination number for shale [unitless].
    shear_red_sst : float
        Shear reduction factor for sandstone [fraction].
    shear_red_mud : float
        Shear reduction factor for mud [fraction].

    Returns
    -------
    tuple
        vp, vs, rho, ai, vpvs  : np.ndarray
        vp [m/s] and vs [m/s], bulk density [kg/m^3], ai [m/s x kg/m^3], vpvs [ratio] of saturated rock.
    """

    # Filter out values of phi that are above phi_c, assumed only to apply for the sandstone
    (
        idx_phi,
        (
            k_sst,
            mu_sst,
            rho_sst,
            k_mud,
            mu_mud,
            rho_mud,
            k_fl_sst,
            rho_fl_sst,
            k_fl_mud,
            rho_fl_mud,
            phi,
            p_eff_sst,
            p_eff_mud,
            shale_frac,
            _,
            _,
        ),
    ) = cast(
        tuple[npt.NDArray[np.bool_], list[npt.NDArray[np.float64]]],
        gen_utilities.filter_input_log(
            (
                k_sst,
                mu_sst,
                rho_sst,
                k_mud,
                mu_mud,
                rho_mud,
                k_fl_sst,
                rho_fl_sst,
                k_fl_mud,
                rho_fl_mud,
                phi,
                p_eff_sst,
                p_eff_mud,
                shale_frac,
                phi_c_sst - phi,
                phi - phi_intr_mud,
            ),
            no_zero=False,
        ),
    )

    # Expand the needed variables from float to numpy array
    phi, phi_intr_mud_ = cast(
        list[npt.NDArray[np.float64]],
        gen_utilities.dim_check_vector((phi, phi_intr_mud)),
    )

    sandy_shale_idx = shale_frac > phi
    shaly_sand_idx = ~sandy_shale_idx

    # Fraction of silt in silt - shale trend
    frac_silt = (1 - shale_frac) / (1 - phi)
    # Fraction of sand in sand - silt trend
    frac_sand = 1 - shale_frac / phi

    # Shale properties for intrinsic porosity point
    vp_sat_mud, vs_sat_mud, rho_b_mud, _, _ = friable_model(
        k_min=k_mud,
        mu_min=mu_mud,
        rho_min=rho_mud,
        k_fl=k_fl_mud,
        rho_fl=rho_fl_mud,
        phi=phi_intr_mud_,
        p_eff=p_eff_mud,
        phi_c=phi_c_mud,
        coord_num_func=coord_num_func_mud,
        n=n_mud,
        shear_red=shear_red_mud,
    )
    k_sat_mud, mu_sat_mud = std_functions.moduli(
        vp=vp_sat_mud,
        vs=vs_sat_mud,
        rhob=rho_b_mud,
    )

    # Silt end member
    k_silt, mu_silt = std_functions.hashin_shtrikman_walpole(
        k1=k_sat_mud,
        mu1=mu_sat_mud,
        k2=k_sst,
        mu2=mu_sst,
        f1=phi_c_sst,
    )
    rho_silt = rho_b_mud * phi + rho_sst * (1 - phi)

    # Estimate the sand end member through the friable model with phi = phiC
    vp_sat_sst, vs_sat_sst, rho_sat_sst, _, _ = friable_model(
        k_min=k_sst,
        mu_min=mu_sst,
        rho_min=rho_sst,
        k_fl=k_fl_sst,
        rho_fl=rho_fl_sst,
        phi=phi,
        p_eff=p_eff_sst,
        phi_c=phi_c_sst,
        coord_num_func=coord_num_func_sst,
        n=n_sst,
        shear_red=shear_red_sst,
    )
    k_sat_sst, mu_sat_sst = std_functions.moduli(
        vp=vp_sat_sst,
        vs=vs_sat_sst,
        rhob=rho_sat_sst,
    )

    k = np.ones(shale_frac.shape) * np.nan
    mu = np.ones(shale_frac.shape) * np.nan
    rho = np.ones(shale_frac.shape) * np.nan

    # Points on sandy shale trend
    k[sandy_shale_idx], mu[sandy_shale_idx] = std_functions.hashin_shtrikman_walpole(
        k1=k_silt[sandy_shale_idx],
        mu1=mu_silt[sandy_shale_idx],
        k2=k_sat_mud[sandy_shale_idx],
        mu2=mu_sat_mud[sandy_shale_idx],
        f1=frac_silt[sandy_shale_idx],
    )

    rho[sandy_shale_idx] = (
        rho_b_mud[sandy_shale_idx] * (1 - frac_silt[sandy_shale_idx])
        + rho_silt[sandy_shale_idx] * frac_silt[sandy_shale_idx]
    )

    # Points on shaly sand trend
    k[shaly_sand_idx], mu[shaly_sand_idx] = std_functions.hashin_shtrikman_walpole(
        k1=k_sat_sst[shaly_sand_idx],
        mu1=mu_sat_sst[shaly_sand_idx],
        k2=k_silt[shaly_sand_idx],
        mu2=mu_silt[shaly_sand_idx],
        f1=frac_sand[shaly_sand_idx],
    )

    rho[shaly_sand_idx] = (
        (1 - phi[shaly_sand_idx]) * rho_sst[shaly_sand_idx]
        + phi[shaly_sand_idx] * rho_fl_sst[shaly_sand_idx]
    ) * frac_sand[shaly_sand_idx] + (1 - frac_sand[shaly_sand_idx]) * rho_silt[
        shaly_sand_idx
    ]

    vp, vs, ai, vpvs = std_functions.velocity(k=k, mu=mu, rhob=rho)

    # Restore original array length
    vp, vs, rho, ai, vpvs = cast(
        list[npt.NDArray[np.float64]],
        gen_utilities.filter_output(idx_phi, (vp, vs, rho, ai, vpvs)),
    )

    return vp, vs, rho, ai, vpvs



import os

import numpy as np

from rock_physics_open.equinor_utilities.snapshot_test_utilities import (
    INITIATE,
    compare_snapshots,
    get_snapshot_name,
    read_snapshot,
    store_snapshot,
)
from rock_physics_open.shale_models import (
    dem_model,
    dem_model_dual_por,
    shale_4_min_dem_overlay,
    shale_model_4_mineral_dem,
)

k1 = 36.8e9 * np.ones(20)
mu1 = 44.0e9 * np.ones(20)
rho1 = 2650 * np.ones(20)
k2 = 15.0e9 * np.ones(20)
mu2 = 7.5e9 * np.ones(20)
rho2 = 2680 * np.ones(20)
k3 = 71.0e9 * np.ones(20)
mu3 = 32.5e9 * np.ones(20)
rho3 = 2720 * np.ones(20)
k4 = 94.9e9 * np.ones(20)
mu4 = 45.0e9 * np.ones(20)
rho4 = 2870 * np.ones(20)
k_fl = 2.7e9 * np.ones(20)
rho_fl = 1005 * np.ones(20)
phi = np.linspace(0.0, 0.20, 20)
frac2 = np.linspace(0.0, 1.0, 20)
asp_1 = 0.85 * np.ones(20)
asp_2 = 0.15 * np.ones(20)
f1 = 0.5 * np.linspace(0.0, 1.0, 20)
f2 = 0.25 * np.linspace(1.0, 0.0, 20)
f3 = 0.25 * np.linspace(1.0, 0.0, 20)
frac_inc = np.linspace(0.0, 0.4, 20)
frac_inc_1 = np.linspace(0.3, 0.2, 20)
asp1 = 0.5 * np.ones(20)
rhob_inp = 2700 * np.ones(20)
asp2 = 0.85 * np.ones(20)
asp3 = 0.25 * np.ones(20)
asp4 = 0.05 * np.ones(20)
asp = 0.85 * np.ones(20)
tol = 1.0e-6
prop_clay = 0.33


def test_dem_model():
    args = dem_model(k1, mu1, rho1, k2, mu2, rho2, frac2, asp_2, tol)

    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), *args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))


def test_dem_dual_por_model():
    args = dem_model_dual_por(
        k1,
        mu1,
        rho1,
        k2,
        mu2,
        rho2,
        k3,
        mu3,
        rho3,
        frac_inc,
        frac_inc_1,
        asp_1,
        asp_2,
        tol,
    )

    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), *args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))


def test_dem_4_min_model():
    args = shale_model_4_mineral_dem(
        k1,
        mu1,
        rho1,
        k2,
        mu2,
        rho2,
        k3,
        mu3,
        rho3,
        k4,
        mu4,
        rho4,
        k_fl,
        rho_fl,
        phi,
        f1,
        f2,
        f3,
        rhob_inp,
        asp1,
        asp2,
        asp3,
        asp4,
        asp,
        mod_type="SCA",
    )

    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), *args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))


def test_dem_4_min_overlay_model():
    args = shale_4_min_dem_overlay(
        k1,
        mu1,
        rho1,
        k2,
        mu2,
        rho2,
        k3,
        mu3,
        rho3,
        k4,
        mu4,
        rho4,
        k_fl,
        rho_fl,
        phi,
        f1,
        f2,
        prop_clay,
        asp,
    )

    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), *args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))



from typing import Literal, cast, overload

import numpy as np
import numpy.typing as npt
from scipy.interpolate import interp1d

from rock_physics_open.equinor_utilities import gen_utilities, std_functions


def constant_cement_model(
    k_min: npt.NDArray[np.float64],
    mu_min: npt.NDArray[np.float64],
    rho_min: npt.NDArray[np.float64],
    k_cem: npt.NDArray[np.float64],
    mu_cem: npt.NDArray[np.float64],
    rho_cem: npt.NDArray[np.float64],
    k_fl: npt.NDArray[np.float64],
    rho_fl: npt.NDArray[np.float64],
    phi: npt.NDArray[np.float64],
    frac_cem: npt.NDArray[np.float64] | float,
    phi_c: float,
    n: float,
    shear_red: float,
    extrapolate_to_max_phi: bool = False,
) -> tuple[
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
]:
    """
    Constant cement model is a sandstone model that combined a cemented and a friable sand, so that a constant
    proportion of the rock volume is made up of grain-bonding cement. Variation in porosity is due to grain sorting,
    i.e. a well-sorted sand will have high porosity, and a poorly sorted one will have low porosity. In the extreme
    end all porosity is removed and the effective mineral properties is returned.

    Mineral properties k_min, mu_min, rho_min are effective properties. For mixtures of minerals, effective
    properties are calculated by Hashin-Shtrikman or similar.

    Cement properties k_cem, mu_cem, rho_cem and cement fraction, the latter as a part of the whole volume.

    Fluid properties k_fl, rho_fl are in situ properties calculated by fluid models using reservoir properties.

    Critical porosity phi_c is input to Dvorkin-Nur function. Coordination number n is normally set to a fixed
    value (default 9) but it is possible to  override this. Porosity phi is used in Hashin-Shtrikman mixing
    (together with phi_c) and Gassmann saturation.

    Shear reduction parameter shear_red is used to account for tangential frictionless grain contacts

    All inputs are assumed to be vectors of the same length

    Parameters
    ----------
    k_min : np.ndarray
        Mineral bulk modulus [Pa].
    mu_min : np.ndarray
        Mineral shear modulus [Pa].
    rho_min : np.ndarray
        Mineral bulk density [kg/m^3].
    k_cem : np.ndarray
        Cement bulk modulus [Pa].
    mu_cem : np.ndarray
        Cement shear modulus [Pa].
    rho_cem : np.ndarray
        Cement bulk density [kg/m^3].
    k_fl : np.ndarray
        Fluid bulk modulus [Pa].
    rho_fl : np.ndarray
        Fluid bulk density [kg/m^3].
    phi : np.ndarray
        Porosity [fraction].
    frac_cem : np.ndarray or float
        Cement fraction [fraction].
    phi_c : float
        Critical porosity [fraction].
    n : float
        Coordination number [unitless].
    shear_red : float
        Shear reduction factor [fraction].
    extrapolate_to_max_phi : bool
        If True, the model will extrapolate to the maximum porosity value (phi_c - frac_cem) if the input porosity
        exceeds this value. If False, the model will return NaN for porosity values exceeding this limit.

    Returns
    -------
    tuple
        vp, vs, rho, ai, vpvs  : np.ndarray
        vp [m/s] and vs [m/s], bulk density [kg/m^3], ai [m/s x kg/m^3], vpvs [ratio] of saturated rock.
    """
    k_zero, k_dry, mu = constant_cement_model_dry(
        k_min=k_min,
        mu_min=mu_min,
        k_cem=k_cem,
        mu_cem=mu_cem,
        phi=phi,
        frac_cem=frac_cem,
        phi_c=phi_c,
        n=n,
        shear_red=shear_red,
        extrapolate_to_max_phi=extrapolate_to_max_phi,
        return_k_zero=True,
    )
    # Saturated rock incompressibility is calculated with Gassmann
    k = std_functions.gassmann(
        k_dry=k_dry,
        por=phi,
        k_fl=k_fl,
        k_min=k_zero,
    )

    # Bulk density
    rho = phi * rho_fl + (1 - phi - frac_cem) * rho_min + frac_cem * rho_cem
    # Velocity
    vp, vs, ai, vpvs = std_functions.velocity(
        k=k,
        mu=mu,
        rhob=rho,
    )

    return vp, vs, rho, ai, vpvs


@overload
def constant_cement_model_dry(
    k_min: npt.NDArray[np.float64],
    mu_min: npt.NDArray[np.float64],
    k_cem: npt.NDArray[np.float64],
    mu_cem: npt.NDArray[np.float64],
    phi: npt.NDArray[np.float64],
    frac_cem: npt.NDArray[np.float64] | float,
    phi_c: float,
    n: float,
    shear_red: npt.NDArray[np.float64] | float,
    extrapolate_to_max_phi: bool,
    return_k_zero: Literal[False],
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]: ...


@overload
def constant_cement_model_dry(
    k_min: npt.NDArray[np.float64],
    mu_min: npt.NDArray[np.float64],
    k_cem: npt.NDArray[np.float64],
    mu_cem: npt.NDArray[np.float64],
    phi: npt.NDArray[np.float64],
    frac_cem: npt.NDArray[np.float64] | float,
    phi_c: float,
    n: float,
    shear_red: npt.NDArray[np.float64] | float,
    extrapolate_to_max_phi: bool = ...,
    return_k_zero: Literal[True] = ...,
) -> tuple[
    npt.NDArray[np.float64], npt.NDArray[np.float64], npt.NDArray[np.float64]
]: ...


def constant_cement_model_dry(
    k_min: npt.NDArray[np.float64],
    mu_min: npt.NDArray[np.float64],
    k_cem: npt.NDArray[np.float64],
    mu_cem: npt.NDArray[np.float64],
    phi: npt.NDArray[np.float64],
    frac_cem: npt.NDArray[np.float64] | float,
    phi_c: float,
    n: float,
    shear_red: npt.NDArray[np.float64] | float,
    extrapolate_to_max_phi: bool = False,
    return_k_zero: bool = False,
) -> (
    tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]
    | tuple[npt.NDArray[np.float64], npt.NDArray[np.float64], npt.NDArray[np.float64]]
):
    """
    Dry rock version of the constant cement model. The method is identical to the constant cement model function,
    except that a saturation step is not performed at the end.

    Parameters
    ----------
    k_min : np.ndarray
        Mineral bulk modulus [Pa].
    mu_min : np.ndarray
        Mineral shear modulus [Pa].
    k_cem : np.ndarray
        Cement bulk modulus [Pa].
    mu_cem : np.ndarray
        Cement shear modulus [Pa].
    phi : np.ndarray
        Porosity [fraction].
    frac_cem : float or np.ndarray
        Cement fraction [fraction].
    phi_c : float
        Critical porosity [fraction].
    n : float
        Coordination number [unitless].
    shear_red : float
        Shear reduction factor [fraction].
    extrapolate_to_max_phi : bool
        If True, the model will extrapolate to the maximum porosity value (phi_c - frac_cem) if the input porosity
        exceeds this value. If False, the model will return NaN for porosity values exceeding this limit.
    return_k_zero : bool
        If True, the model will return the zero-porosity end member bulk modulus k_zero in addition to the dry rock
        bulk modulus k_dry and shear modulus mu.

    Returns
    -------
    tuple
        k_dry, mu : np.ndarray
        Bulk modulus k [Pa] and shear modulus mu [Pa] of dry rock.
    """
    # First check if there are input values that are unphysical, i.e. negative values, separate between dry and
    # saturated rock properties. Use the filter_input_log function to identify these values
    idx, (k_min, mu_min, k_cem, mu_cem, phi) = cast(
        tuple[npt.NDArray[np.bool_], list[npt.NDArray[np.float64]]],
        gen_utilities.filter_input_log((k_min, mu_min, k_cem, mu_cem, phi)),
    )

    # Identify porosity values that exceed (phi_c - frac_cem). This is regardless of setting for extrapolate_to_max_phi
    idx_phi = np.where(phi > phi_c - frac_cem)[0]

    # At the zero-porosity point, all original porosity (critical porosity -
    # cement fraction) is filled with grains. The cement fraction surrounds the
    # original grains, so they will be fraction 1 according to the geometrical
    # interpretation of Hashin-Shtrikman
    k_zero, mu_zero = std_functions.hashin_shtrikman_walpole(
        k1=k_cem,
        mu1=mu_cem,
        k2=k_min,
        mu2=mu_min,
        f1=frac_cem,
        bound="lower",
    )

    # Dry rock properties of high-porosity end member calculated with
    # Dvorkin-Nur equation. Cement is assumed to be evenly distributed on the
    # grains (scheme 2 in Dvorkin and Nur's original paper)

    k_cc, mu_cc = std_functions.dvorkin_contact_cement(
        frac_cem=frac_cem,
        por0_sst=phi_c,
        mu0_sst=mu_min,
        k0_sst=k_min,
        mu0_cem=mu_cem,
        k0_cem=k_cem,
        vs_red=shear_red,
        c=n,
    )

    # Hashin-Shtrikman lower bound describes the dry rock property mixing from
    # mineral properties to high-end porosity.

    # Fraction of zero-porosity end member
    f1 = 1 - phi / (phi_c - frac_cem)

    k_dry, mu = std_functions.hashin_shtrikman_walpole(
        k1=k_zero,
        mu1=mu_zero,
        k2=k_cc,
        mu2=mu_cc,
        f1=f1,
        bound="lower",
    )

    # If extrapolate_to_max_phi is True, create extrapolation functions for k_dry and mu using scipy's interp1d
    if extrapolate_to_max_phi:
        k_dry_func = interp1d(phi, k_dry, fill_value="extrapolate")
        mu_func = interp1d(phi, mu, fill_value="extrapolate")
        k_dry[idx_phi] = k_dry_func(phi[idx_phi])
        mu[idx_phi] = mu_func(phi[idx_phi])
    else:
        k_dry[idx_phi] = np.nan
        mu[idx_phi] = np.nan

    k_dry, mu = cast(
        list[npt.NDArray[np.float64]],
        gen_utilities.filter_output(idx, (k_dry, mu)),
    )
    if return_k_zero:
        return k_zero, k_dry, mu
    return k_dry, mu



from typing import Literal, cast

import numpy as np
import numpy.typing as npt

from rock_physics_open.equinor_utilities.gen_utilities import dim_check_vector


def hashin_shtrikman(
    k1: npt.NDArray[np.float64],
    mu1: npt.NDArray[np.float64],
    k2: npt.NDArray[np.float64],
    mu2: npt.NDArray[np.float64],
    f: npt.NDArray[np.float64],
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    Hashin-Sktrikman upper or lower according to ordering of phases.

    Parameters
    ----------
    k1 : np.ndarray
        Bulk modulus of phase 1 [Pa].
    mu1 : np.ndarray
        Shear modulus of phase 1 [Pa].
    k2 : np.ndarray
        Bulk modulus of phase 2 [Pa].
    mu2 : np.ndarray
        Shear modulus of phase 2 [Pa].
    f : np.ndarray
        Fraction of phase 1 [fraction].

    Returns
    -------
    tuple
        k, mu : np.ndarray.
        k: effective bulk modulus [Pa], mu: effective shear modulus [Pa].
    """
    k = k1 + (1 - f) * (k2 - k1) / (1 + (k2 - k1) * f * (k1 + 4 / 3 * mu1) ** -1)
    mu = mu1 + (1 - f) * (mu2 - mu1) / (
        1 + 2 * (mu2 - mu1) * f * (k1 + 2 * mu1) / (5 * mu1 * (k1 + 4 / 3 * mu1))
    )

    return k, mu


def hashin_shtrikman_average(
    k1: npt.NDArray[np.float64],
    mu1: npt.NDArray[np.float64],
    k2: npt.NDArray[np.float64],
    mu2: npt.NDArray[np.float64],
    f: npt.NDArray[np.float64],
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    Average of Hashin-Shtrikman upper and lower bound.

    Parameters
    ----------
    k1 : np.ndarray
        Bulk modulus of phase 1 [Pa].
    mu1 : np.ndarray
        Shear modulus of phase 1 [Pa].
    k2 : np.ndarray
        Bulk modulus of phase 2 [Pa].
    mu2 : np.ndarray
        Shear modulus of phase 2 [Pa].
    f : np.ndarray
        Fraction of phase 1 [fraction].

    Returns
    -------
    tuple
        k_av, mu_av : np.ndarray.
        k_av: effective bulk modulus [Pa], mu_av: effective shear modulus [Pa]
    """
    k_hs1, mu_hs1 = hashin_shtrikman(k1, mu1, k2, mu2, f)
    k_hs2, mu_hs2 = hashin_shtrikman(k2, mu2, k1, mu1, 1 - f)

    k_av = (k_hs1 + k_hs2) / 2
    mu_av = (mu_hs1 + mu_hs2) / 2

    return k_av, mu_av


def hashin_shtrikman_walpole(
    k1: npt.NDArray[np.float64],
    mu1: npt.NDArray[np.float64],
    k2: npt.NDArray[np.float64],
    mu2: npt.NDArray[np.float64],
    f1: npt.NDArray[np.float64] | float,
    bound: Literal["upper", "lower"] = "lower",
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    Hashin-Shtrikman upper bound is obtained when the stiffest material is
    termed 1 and vice versa for lower bound. Tricky in cases like Quartz -
    Calcite where the K and Mu have opposed values. HS - Walpole is
    generalised to regard highest and lowest values in each case. The default
    is to generate lower bound.

    Parameters
    ----------
    k1 : np.ndarray
        Bulk modulus of phase 1 [Pa].
    mu1 : np.ndarray
        Shear modulus of phase 1 [Pa].
    k2 : np.ndarray
        Bulk modulus of phase 2 [Pa].
    mu2 : np.ndarray
        Shear modulus of phase 2 [Pa].
    f1 : np.ndarray or float
        Fraction of phase 1 [fraction].
    bound: str
        'upper' or 'lower' selection of upper of lower bound of effective medium.
    Returns
    -------
    tuple
        k, mu : np.ndarray.
        k: effective bulk modulus [Pa], mu: effective shear modulus [Pa].
    """
    k1, mu1, k2, mu2, f1 = cast(
        list[npt.NDArray[np.float64]],
        dim_check_vector((k1, mu1, k2, mu2, f1)),
    )
    if bound.lower() not in ["lower", "upper"]:
        raise ValueError(f'{__file__}: bound must be one of "lower" or "upper"')

    idx_k = k1 == k2
    idx_mu = mu1 == mu2
    f2 = 1 - f1

    if bound.lower() == "lower":
        k_m = np.minimum(k1, k2)
        mu_m = np.minimum(mu1, mu2)
    else:
        k_m = np.maximum(k1, k2)
        mu_m = np.maximum(mu1, mu2)

    k = np.zeros(k1.shape)
    mu = np.zeros(k1.shape)

    if np.any(idx_k):
        k[idx_k] = k1[idx_k]
    if np.any(~idx_k):
        k[~idx_k] = k1[~idx_k] + f2[~idx_k] / (
            (k2[~idx_k] - k1[~idx_k]) ** -1
            + f1[~idx_k] * (k1[~idx_k] + 4 / 3 * mu_m[~idx_k]) ** -1
        )
    if np.any(idx_mu):
        mu[idx_mu] = mu1[idx_mu]
    if np.any(~idx_mu):
        mu[~idx_mu] = mu1[~idx_mu] + f2[~idx_mu] / (
            (mu2[~idx_mu] - mu1[~idx_mu]) ** -1
            + f1[~idx_mu]
            * (
                mu1[~idx_mu]
                + mu_m[~idx_mu]
                / 6
                * (9 * k_m[~idx_mu] + 8 * mu_m[~idx_mu])
                / (k_m[~idx_mu] + 2 * mu_m[~idx_mu])
            )
            ** -1
        )

    return k, mu


def multi_hashin_shtrikman(
    *coeffs: npt.NDArray[np.float64],
    mode: Literal["average", "upper", "lower"] = "average",
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    Hashin-Shtrikman effective medium calculation for multi-mineral case.

    Parameters
    ----------
    coeffs : np.ndarray
        Triplets of vectors with k (bulk modulus [Pa]), mu (shear modulus [Pa]) and fraction for each mineral.
    mode : str
        'average', 'upper' or 'lower'.

    Returns
    -------
    tuple
        k_hs, mu_hs : np.ndarray.
        k_hs, mu_hs - bulk modulus and shear modulus for effective medium [Pa].
    """
    if not len(coeffs) % 3 == 0:
        raise ValueError(
            "multi_hashin_shtrikman: inputs not vectors of k, mu and fraction for each mineral"
        )

    k_arr = np.array(coeffs[::3])
    mu_arr = np.array(coeffs[1::3])
    f = np.array(coeffs[2::3])
    if not np.all(
        cast(npt.NDArray[np.float64], np.around(np.sum(f, axis=0), decimals=6)) == 1.0
    ):
        raise ValueError("multi_hashin_shtrikman: all fractions do not add up to 1.0")

    if mode.lower() not in ["average", "upper", "lower"]:
        raise ValueError(
            'multi_hashin_shtrikman: mode is not one of "average", "upper" or "lower"'
        )

    k_min = np.min(k_arr, axis=0)
    k_max = np.max(k_arr, axis=0)
    mu_min = np.min(mu_arr, axis=0)
    mu_max = np.max(mu_arr, axis=0)

    k_hs_upper = np.sum(f / (k_arr + 4 / 3 * mu_max), axis=0) ** -1 - 4 / 3 * mu_max
    k_hs_lower = np.sum(f / (k_arr + 4 / 3 * mu_min), axis=0) ** -1 - 4 / 3 * mu_min

    zeta_max = mu_max / 6 * (9 * k_max + 8 * mu_max) / (k_max + 2 * mu_max)
    zeta_min = mu_min / 6 * (9 * k_min + 8 * mu_min) / (k_min + 2 * mu_min)

    mu_hs_upper = np.sum(f / (mu_arr + zeta_max), axis=0) ** -1 - zeta_max
    mu_hs_lower = np.sum(f / (mu_arr + zeta_min), axis=0) ** -1 - zeta_min

    if mode == "lower":
        return k_hs_lower, mu_hs_lower
    if mode == "upper":
        return k_hs_upper, mu_hs_upper
    k_hs = 0.5 * (k_hs_upper + k_hs_lower)
    mu_hs = 0.5 * (mu_hs_upper + mu_hs_lower)
    return k_hs, mu_hs



from .compare_snapshots import compare_snapshots
from .snapshots import INITIATE, get_snapshot_name, read_snapshot, store_snapshot

__all__ = [
    "compare_snapshots",
    "INITIATE",
    "get_snapshot_name",
    "read_snapshot",
    "store_snapshot",
]



import os
from collections.abc import Sequence
from pathlib import Path
from re import match
from typing import Any, cast

import numpy as np
import numpy.typing as npt
import pandas as pd
from sklearn.preprocessing import OneHotEncoder, RobustScaler

from .dummy_vars import generate_dummy_vars
from .import_ml_models import AnyModel, import_model


def _read_models(
    *model_files: str, model_dir: str | Path | None = None
) -> tuple[
    list[AnyModel],
    list[RobustScaler],
    list[OneHotEncoder | None],
    list[str],
    list[str],
    list[list[str]],
    list[Any],
    list[str],
    list[str],
]:
    # Find the directory of the model files, change working directory, return to original directory at end of function
    orig_dir = os.getcwd()
    if model_dir is None:
        model_dir, _ = os.path.split(model_files[0])
    os.chdir(model_dir)
    # Allocate lists and read model
    reg_models: list[AnyModel] = []
    scalers: list[RobustScaler] = []
    ohes: list[OneHotEncoder | None] = []
    label_vars: list[str] = []
    label_units: list[str] = []
    feat_vars: list[list[str]] = []
    cat_vars: list[Any] = []
    col_names: list[str] = []
    col_units: list[str] = []

    for mod_name in model_files:
        models, scaler, ohe, label_var, label_unit, feat_var, cat_var = import_model(
            model_file_name=mod_name
        )
        reg_models.append(models)
        scalers.append(scaler)
        ohes.append(ohe)
        label_vars.append(label_var)
        # Need to modify names
        col_names.append(label_var + "_" + mod_name.replace(label_var, ""))
        col_units.append(label_unit)
        label_units.append(label_unit)
        feat_vars.append(feat_var)
        cat_vars.append(cat_var)

        # Need to modify names
        col_names, col_units = ([] for _ in range(2))
        for i in range(len(label_vars)):
            col_names.append(
                label_vars[i] + "_" + model_files[i].replace(label_vars[i], "")
            )
            col_units.append(label_units[i])

    os.chdir(orig_dir)

    return (
        reg_models,
        scalers,
        ohes,
        label_vars,
        label_units,
        feat_vars,
        cat_vars,
        col_names,
        col_units,
    )


def _perform_regression(
    inp_frame: pd.DataFrame,
    col_names: list[str],
    feat_var: list[list[str]],
    cat_var: str | list[str],
    ohe: Sequence[OneHotEncoder | None],
    scaler: Sequence[RobustScaler | None],
    reg_model: Sequence[AnyModel],
) -> pd.DataFrame:
    depth = inp_frame.index.to_numpy()

    res_frame = pd.DataFrame(index=depth, columns=col_names)

    for j, _ in enumerate(col_names):
        tmp_frame = inp_frame.copy()

        # Limit to columns used in estimation before dropping NaNs
        num_var = [i for i in feat_var[j] if not bool(match(r"x\d", i))]
        no_num_var = len(num_var)
        if cat_var[j]:
            num_var.append(cat_var[j])
        tmp_frame = tmp_frame[num_var]
        idx_na_n = tmp_frame.isna().any(axis=1)

        if cat_var[j]:
            dum_features, _, dum_var_names = generate_dummy_vars(
                inp_frame=tmp_frame.loc[~idx_na_n],
                class_var=cat_var[j],
                ohe=ohe[j],
            )
            # Add dummy features to data frame
            kept_dum_var: list[str] = []
            for i, name in enumerate(dum_var_names):
                if name in feat_var[j]:
                    tmp_frame.loc[~idx_na_n, name] = dum_features[:, i]
                    kept_dum_var.append(name)
            tmp_frame.drop(columns=[cat_var[j]], inplace=True)

            # Need to assure that we have the correct sequence of features
            tmp_frame = tmp_frame.reindex(columns=feat_var[j])

            new_features = np.zeros((np.sum(~idx_na_n), tmp_frame.shape[1]))
            # Make scaling optional
            scaler_ = scaler[j]
            if scaler_ is not None:
                new_features = (
                    cast(  # Casting since scikit-learn is not yet fully typed.
                        npt.NDArray[np.float64],
                        scaler_.transform(tmp_frame.to_numpy()[~idx_na_n, :]),
                    )
                )
            else:
                new_features[:, :no_num_var] = tmp_frame.to_numpy()[
                    ~idx_na_n, :no_num_var
                ]
            new_features[:, no_num_var:] = tmp_frame.loc[
                ~idx_na_n, kept_dum_var
            ].to_numpy()
        else:
            # Much simpler if there are no dummy variables
            # Need to assure that we have the correct sequence of features
            tmp_frame = tmp_frame.reindex(columns=feat_var[j])
            # Make scaling optional
            scaler_ = scaler[j]
            if scaler_ is not None:
                new_features = (
                    cast(  # Casting since scikit-learn is not yet fully typed.
                        npt.NDArray[np.float64],
                        scaler_.transform(tmp_frame.to_numpy()[~idx_na_n, :]),
                    )
                )
            else:
                new_features = tmp_frame.to_numpy()[~idx_na_n, :]

        new_var = np.ones(depth.shape[0]) * np.nan
        new_var[~idx_na_n] = reg_model[j].predict(new_features).flatten()
        res_frame[col_names[j]] = new_var

    return res_frame


def run_regression(
    inp_df: pd.DataFrame,
    first_model_file_name: str,
    second_model_file_name: str,
    model_dir: str | None = None,
) -> pd.DataFrame:
    """
    Estimate Vp and Vs by neural network regression with multiple inputs.

    Parameters
    ----------
    inp_df : pd.DataFrame
        Input logs required for the regression.
    first_model_file_name : str
        Full file name for vp model.
    second_model_file_name : str
        Full file name for vs model.
    model_dir : str | None
        Directory.

    Returns
    -------
    vp, vs : pd.DataFrame
        Estimated vp and vs as series in Pandas DataFrame.
    """

    (
        regression_model,
        scaler_obj,
        ohe_obj,
        _,
        _,
        feature_var,
        category_var,
        column_names,
        _,
    ) = _read_models(first_model_file_name, second_model_file_name, model_dir=model_dir)
    return _perform_regression(
        inp_frame=inp_df,
        col_names=column_names,
        feat_var=feature_var,
        cat_var=category_var,
        ohe=ohe_obj,
        scaler=scaler_obj,
        reg_model=regression_model,
    )



import os
from typing import Literal, cast

import numpy as np
import numpy.typing as npt
import pandas as pd
from sklearn.metrics import r2_score

from .display_result_statistics import disp_result_stats


def vp_vs_rho_stats(
    vp_observed: npt.NDArray[np.float64] | list[npt.NDArray[np.float64]],
    vs_observed: npt.NDArray[np.float64] | list[npt.NDArray[np.float64]],
    rho_observed: npt.NDArray[np.float64] | list[npt.NDArray[np.float64]],
    vp_estimated: npt.NDArray[np.float64] | list[npt.NDArray[np.float64]],
    vs_estimated: npt.NDArray[np.float64] | list[npt.NDArray[np.float64]],
    rho_estimated: npt.NDArray[np.float64] | list[npt.NDArray[np.float64]],
    fname: str,
    estimated_set_names: str | list[str],
    well_names: str | list[str],
    file_mode: Literal["a", "w"] = "a",
    disp_results: bool = True,
) -> None:
    """
    Utility to estimate statistics between vp-vs-rho sets - observed and estimated values. The results are displayed
    on screen (optional) and saved to a .csv file. If the file exists, the results will be appended.
    The statistics consists of RAE (relative absolute error) [fraction], RRMSE (relative root mean squared error)
    [fraction] and R^2 (coefficient of determination) [fraction].


    Parameters
    ----------
    vp_observed : np.ndarray or list
        Observed vp [m/s].
    vs_observed : np.ndarray or list
        Observed vs [m/s].
    rho_observed : np.ndarray or list
        Observed density [kg/m^3].
    vp_estimated : np.ndarray or list
        Estimated vp [m/s].
    vs_estimated : np.ndarray or list
        Estimated vs [m/s].
    rho_estimated : np.ndarray or list
        Estimated density [kg/m^3].
    fname : str
        File name for saved results.
    estimated_set_names : str or list
        Name of the estimated vp-vs-rho set(s).
    well_names : str or list
        Well name of the vp-vs-rho set(s).
    file_mode : str
        File open mode, default append.
    disp_results : bool
        Display results on screen.
    """
    if isinstance(estimated_set_names, str):
        estimated_set_names = [estimated_set_names]
    if isinstance(well_names, str):
        well_names = [well_names]

    _verify(
        vp_observed,
        vs_observed,
        rho_observed,
        vp_estimated,
        vs_estimated,
        rho_estimated,
        set_names=estimated_set_names,
        well_names=well_names,
        file_mode=file_mode,
    )

    est_frame_columns = [
        "Well",
        "Vp RMAE",
        "Vp RRMSE",
        "Vp R2",
        "Vs RMAE",
        "Vs RRMSE",
        "Vs R2",
        "Rho RMAE",
        "Rho RRMSE",
        "Rho R2",
    ]
    est_frame = pd.DataFrame(columns=est_frame_columns, index=estimated_set_names)
    est_frame.index.name = "Estimated set name"
    est_frame.iloc[:, 0] = well_names

    # If inputs are found to satisfy expectations in _verify, and they are numpy arrays, cast to lists, and run through
    if isinstance(vp_observed, np.ndarray):
        vp_observed = [vp_observed]
    if isinstance(vs_observed, np.ndarray):
        vs_observed = [vs_observed]
    if isinstance(rho_observed, np.ndarray):
        rho_observed = [rho_observed]
    if isinstance(vp_estimated, np.ndarray):
        vp_estimated = [vp_estimated]
    if isinstance(vs_estimated, np.ndarray):
        vs_estimated = [vs_estimated]
    if isinstance(rho_estimated, np.ndarray):
        rho_estimated = [rho_estimated]

    for i in range(len(vp_observed)):
        res: list[float] = []
        for obs, est in zip(
            [vp_observed[i], vs_observed[i], rho_observed[i]],
            [vp_estimated[i], vs_estimated[i], rho_estimated[i]],
        ):
            rmae = float(
                np.mean(np.abs((est.flatten() - obs.flatten()) / obs.flatten()))
            )
            rrmse: float = np.sqrt(
                np.mean(np.square((est.flatten() - obs.flatten()) / obs.flatten()))
            )
            r2 = cast(float, r2_score(obs.flatten(), est.flatten()))

            res.append(rmae)
            res.append(rrmse)
            res.append(r2)

        res_dict = dict(zip(est_frame_columns[1:], res))
        est_frame.iloc[i, 1:] = res_dict  # pyright: ignore[reportArgumentType]
        if disp_results:
            disp_result_stats(
                estimated_set_names[i], res, est_frame_columns[1:], values_only=True
            )
    # Test if the file already exists. If so, and the file_mode is set to 'a', drop the column headers when writing
    # the file
    if os.path.exists(fname) and file_mode == "a":
        est_frame.to_csv(fname, mode=file_mode, header=False)
    else:
        est_frame.to_csv(fname, mode=file_mode)


def _verify(
    *args: npt.NDArray[np.float64] | list[npt.NDArray[np.float64]],
    set_names: list[str],
    well_names: list[str],
    file_mode: Literal["a", "w"],
):
    """Verify that arguments are either numpy arrays or lists of numpy arrays.
    Raises
    ------
    ValueError
        If not np.ndarray or list of such.
    ValueError
        If an entry contains NaNs.
    ValueError
        If an entry contains Infs.
    ValueError
        If mismatch in argument lengths.
    ValueError
        If file mode not 'a' or 'w'.
    """
    # Verify that args are either numpy arrays or list of arrays
    for arg in args:
        if isinstance(arg, np.ndarray):
            arg = [arg]
        for this_arg in arg:
            if not isinstance(this_arg, np.ndarray):  # pyright: ignore[reportUnnecessaryIsInstance] | For backward compatibility
                raise ValueError(f"{__file__}: input not numpy array: {type(arg)}")
            if np.any(np.isnan(this_arg)):
                raise ValueError(f"{__file__}: input contains NaNs")
            if np.any(np.isinf(this_arg)):
                raise ValueError(f"{__file__}: input contains Infs")
        if not len(arg) == len(set_names) == len(well_names):
            raise ValueError(f"{__file__}: mismatch in argument lengths")
    if not (file_mode == "a" or file_mode == "w"):
        raise ValueError(f'{__file__}: file_mode must be one of ["a", "w"]')  # pyright: ignore[reportUnreachable] | For backward compatibility



"""
Simple conversions required for the material models.
"""

import numpy as np
import numpy.typing as npt


def celsius_to_kelvin(
    temperature: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    Convert temperature from Celsius to kelvin
    """
    return temperature + 273.15



from typing import cast

import numpy as np
import numpy.typing as npt
from matplotlib import path as pa, pyplot as plt
from matplotlib.colorbar import Colorbar
from matplotlib.text import Text

from rock_physics_open.equinor_utilities.various_utilities.types import Array2D

from .ternary_plot_utilities import (
    make_mesh,
    set_ternary_figure,
    ternary_coord_trans,
)


def ternary_patches(
    quartz: npt.NDArray[np.float64],
    carb: npt.NDArray[np.float64],
    clay: npt.NDArray[np.float64],
    kero: npt.NDArray[np.float64],
    well_name: str,
    draw_figures: bool = True,
) -> npt.NDArray[np.float64]:
    """Class display of the shale.

    Parameters
    ----------
    quartz : np.ndarray
        Quartz volume fraction [fraction].
    carb : np.ndarray
        Carbonate volume fraction [fraction].
    clay : np.ndarray
        Clay volume fraction [fraction].
    kero : np.ndarray
        Kerogen volume fraction [fraction].
    well_name : str
        Plot heading with well name.
    draw_figures : bool
        Decide if figures are drawn or not, default is True.

    Returns
    -------
    np.ndarray
        Lithology class [int].
    """
    _, ax = set_ternary_figure(
        delta_x=0,
        delta_y=0,
        title="Lithological Class Ternary Plot",
        well_name=well_name,
    )

    # Define patches 1 - 16
    names = [
        "Silica-\ndominated\nlithotype",
        "Clay-rich\nsiliceous mudstone",
        "Mixed\nsiliceous\nmudstone",
        "Carbonate-rich\nsiliceous mudstone",
        "Argilaceous/\nsiliceous\nmudstone",
        "Mixed\nmudstone",
        "Carbonate/\nsiliceous\nmudstone",
        "Silica-rich\nargillaceous mudstone",
        "Mixed\nargillaceous\nmudstone",
        "Argillaceous/\ncarbonate\nmudstone\n(marl)",
        "Mixed\ncarbonate\nmudstone",
        "Silica-rich\ncarbonate mudstone",
        "Clay-\ndominated\nlithotype",
        "Carbonate-rich\nargillaceous mudstone",
        "Clay-rich\ncarbonate mudstone",
        "Carbonate-\ndominated\nlithotype",
    ]
    text_ang = [0, 60, 0, 0, 0, 0, 0, 60, 0, 0, 0, 0, 0, -60, -60, 0]
    col = (
        1
        / 255
        * np.array(
            [
                [244, 237, 29],
                [249, 246, 142],
                [255, 230, 137],
                [249, 246, 137],
                [254, 220, 95],
                [222, 238, 197],
                [131, 209, 193],
                [214, 170, 40],
                [197, 166, 114],
                [169, 169, 134],
                [193, 215, 239],
                [150, 180, 218],
                [175, 116, 76],
                [169, 123, 44],
                [96, 149, 202],
                [114, 126, 188],
            ]
        )
    )

    vertices: list[Array2D[np.float64]] = [
        np.array(
            [
                [1.0, 0.0, 0.0],
                [0.8, 0.0, 0.2],
                [0.8, 0.2, 0.0],
            ]
        ),
        np.array(
            [
                [0.8, 0.0, 0.2],
                [0.5, 0.0, 0.5],
                [0.5, 0.1, 0.4],
                [0.8, 0.1, 0.1],
            ]
        ),
        np.array(
            [
                [0.8, 0.1, 0.1],
                [0.5, 0.1, 0.4],
                [0.5, 0.4, 0.1],
            ]
        ),
        np.array(
            [
                [0.8, 0.1, 0.1],
                [0.5, 0.4, 0.1],
                [0.5, 0.5, 0.0],
                [0.8, 0.2, 0.0],
            ]
        ),
        np.array(
            [
                [0.5, 0.0, 0.5],
                [0.3, 0.2, 0.5],
                [0.5, 0.2, 0.3],
            ]
        ),
        np.array(
            [
                [0.5, 0.2, 0.3],
                [0.3, 0.2, 0.5],
                [0.2, 0.3, 0.5],
                [0.2, 0.5, 0.3],
                [0.3, 0.5, 0.2],
                [0.5, 0.3, 0.2],
            ]
        ),
        np.array(
            [
                [0.5, 0.3, 0.2],
                [0.3, 0.5, 0.2],
                [0.5, 0.5, 0.0],
            ]
        ),
        np.array(
            [
                [0.5, 0.0, 0.5],
                [0.2, 0.0, 0.8],
                [0.1, 0.1, 0.8],
                [0.4, 0.1, 0.5],
            ]
        ),
        np.array(
            [
                [0.4, 0.1, 0.5],
                [0.1, 0.1, 0.8],
                [0.1, 0.4, 0.5],
            ]
        ),
        np.array(
            [
                [0.0, 0.5, 0.5],
                [0.2, 0.5, 0.3],
                [0.2, 0.3, 0.5],
            ]
        ),
        np.array(
            [
                [0.1, 0.5, 0.4],
                [0.1, 0.8, 0.1],
                [0.4, 0.5, 0.1],
            ]
        ),
        np.array(
            [
                [0.4, 0.5, 0.1],
                [0.1, 0.8, 0.1],
                [0.2, 0.8, 0.0],
                [0.5, 0.5, 0.0],
            ]
        ),
        np.array(
            [
                [0.2, 0.0, 0.8],
                [0.0, 0.0, 1.0],
                [0.0, 0.2, 0.8],
            ]
        ),
        np.array(
            [
                [0.0, 0.2, 0.8],
                [0.0, 0.5, 0.5],
                [0.1, 0.4, 0.5],
                [0.1, 0.1, 0.8],
            ]
        ),
        np.array(
            [
                [0.0, 0.5, 0.5],
                [0.0, 0.8, 0.2],
                [0.1, 0.8, 0.1],
                [0.1, 0.5, 0.4],
            ]
        ),
        np.array(
            [
                [0.0, 0.8, 0.2],
                [0.0, 1.0, 0.0],
                [0.2, 0.8, 0.0],
            ]
        ),
    ]

    vertices_xy = [ternary_coord_trans(v) for v in vertices]

    _ = [
        plt.fill(
            vertices_xy[i][:, 0],
            vertices_xy[i][:, 1],
            facecolor=col[i, :],
            edgecolor="k",
        )
        for i in range(len(vertices))
    ]
    _ = [
        plt.text(
            x=float(np.mean(vertices_xy[i][:, 0])),
            y=float(np.mean(vertices_xy[i][:, 1])),
            s=names[i],
            horizontalalignment="center",
            verticalalignment="center",
            fontsize=8,
            rotation=text_ang[i],
        )
        for i in range(len(vertices))
    ]

    # Make coordinate mesh
    _, _ = make_mesh(ax)

    # Annotate the axes
    label_handles: list[Text] = []
    label_handles.append(
        plt.text(
            x=0.5,
            y=-0.075,
            s="Carbonate",
            horizontalalignment="center",
        )
    )
    label_handles.append(
        plt.text(
            x=0.15,
            y=np.sqrt(3) / 4 + 0.05,
            s="Quartz",
            horizontalalignment="center",
            rotation=60,
        )
    )
    label_handles.append(
        plt.text(
            x=0.85,
            y=np.sqrt(3) / 4 + 0.05,
            s="Clay",
            horizontalalignment="center",
            rotation=-60,
        )
    )
    plt.setp(label_handles, fontname="sans-serif", fontweight="bold", fontsize=11)

    # Plot data
    data_xy = ternary_coord_trans(quartz, carb, clay)
    _ = plt.scatter(x=data_xy[:, 0], y=data_xy[:, 1], s=64, c=kero, zorder=4)

    # Classification - start at 2001 to match predefined lithology classes in
    startclass = 2000
    lith_class = np.ones(quartz.shape) * np.nan
    for i in range(len(vertices_xy)):
        # class_path = pa.Path(vertices_xy[i], closed=True)
        class_path = pa.Path(vertices_xy[i])
        in_path = class_path.contains_points(data_xy)
        lith_class[in_path] = i + startclass

    hcb = cast(  # Casting due to incomplete typing in matplotlib
        Colorbar,
        plt.colorbar(),
    )
    hcb.set_label("Kerogen", fontname="sans-serif", fontweight="bold", fontsize=11)

    # plt.show() <= Show command must be issued from the controlling class method
    if draw_figures:
        plt.draw()
    return lith_class



import numpy as np
import numpy.typing as npt

from rock_physics_open.equinor_utilities import std_functions
from rock_physics_open.sandstone_models.friable_models import (
    CoordinateNumberFunction,
    friable_model,
)


def unresolved_friable_sand_shale_model(
    k_sst: npt.NDArray[np.float64],
    mu_sst: npt.NDArray[np.float64],
    rho_sst: npt.NDArray[np.float64],
    k_mud: npt.NDArray[np.float64],
    mu_mud: npt.NDArray[np.float64],
    rho_mud: npt.NDArray[np.float64],
    k_fl_sst: npt.NDArray[np.float64],
    rho_fl_sst: npt.NDArray[np.float64],
    k_fl_mud: npt.NDArray[np.float64],
    rho_fl_mud: npt.NDArray[np.float64],
    phi_sst: npt.NDArray[np.float64],
    phi_mud: npt.NDArray[np.float64],
    p_eff_sst: npt.NDArray[np.float64],
    p_eff_mud: npt.NDArray[np.float64],
    shale_frac: npt.NDArray[np.float64],
    phi_c_sst: float,
    phi_c_mud: float,
    coord_num_func_sst: CoordinateNumberFunction,
    n_sst: float,
    coord_num_func_mud: CoordinateNumberFunction,
    n_mud: float,
    shear_red_sst: float,
    shear_red_mud: float,
) -> tuple[
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
]:
    """
    Model for siliciclastic rocks with alternating layers of friable sand and shale, and in which the layers are not
    resolved by the investigating signal. Backus average is used to calculate the anisotropic effect of the alternating
    layers.

    Parameters
    ----------
    k_sst : np.ndarray
        Sandstone bulk modulus [Pa].
    mu_sst : np.ndarray
        Sandstone shear modulus [Pa].
    rho_sst : np.ndarray
        Sandstone bulk density [kg/m^3].
    k_mud : np.ndarray
        Shale bulk modulus [Pa].
    mu_mud : np.ndarray
        Shale shear modulus [Pa].
    rho_mud : np.ndarray
        Shale bulk density [kg/m^3].
    k_fl_sst : np.ndarray
        Fluid bulk modulus for sandstone fluid [Pa].
    rho_fl_sst : np.ndarray
        Fluid bulk density for sandstone fluid [kg/m^3].
    k_fl_mud : np.ndarray
        Fluid bulk modulus for shale fluid [Pa].
    rho_fl_mud : np.ndarray
        Fluid bulk density for shale fluid [kg/m^3].
    phi_sst : np.ndarray
        Sandstone porosity [fraction].
    phi_mud : np.ndarray
        Shale porosity [fraction].
    p_eff_sst : np.ndarray
        Effective pressure in sandstone [Pa].
    p_eff_mud : np.ndarray
        Effective pressure in mud [Pa].
    shale_frac : np.ndarray
        Shale fraction [fraction].
    phi_c_sst : float
        Critical porosity for sandstone [fraction].
    phi_c_mud : float
        Critical porosity for mud [fraction].
    coord_num_func_sst : str
        Indication if coordination number should be calculated from porosity or kept constant for sandstone.
    n_sst : float
        Coordination number for sandstone [unitless].
    coord_num_func_mud : str
        Indication if coordination number should be calculated from porosity or kept constant for shale.
    n_mud : float
        Coordination number for shale [unitless].
    shear_red_sst : float
        Shear reduction factor for sandstone [fraction].
    shear_red_mud : float
        Shear reduction factor for mud [fraction].

    Returns
    -------
    tuple
        vpv, vsv, vph, vsh, rho : np.ndarray
        vertical p-wave velocity, vertical shear-wave velocity, horizontal p-wave velocity, horizontal shear-wave
        velocity (all [m/s]), bulk density [kg/m^3].
    """

    # Estimate the sand and shale end members through the friable models
    vp_sst, vs_sst, rho_b_sst, _, _ = friable_model(
        k_min=k_sst,
        mu_min=mu_sst,
        rho_min=rho_sst,
        k_fl=k_fl_sst,
        rho_fl=rho_fl_sst,
        phi=phi_sst,
        p_eff=p_eff_sst,
        phi_c=phi_c_sst,
        coord_num_func=coord_num_func_sst,
        n=n_sst,
        shear_red=shear_red_sst,
    )

    vp_mud, vs_mud, rho_b_mud, _, _ = friable_model(
        k_min=k_mud,
        mu_min=mu_mud,
        rho_min=rho_mud,
        k_fl=k_fl_mud,
        rho_fl=rho_fl_mud,
        phi=phi_mud,
        p_eff=p_eff_mud,
        phi_c=phi_c_mud,
        coord_num_func=coord_num_func_mud,
        n=n_mud,
        shear_red=shear_red_mud,
    )

    # Calculate Backus average for the effective medium
    vpv, vsv, vph, vsh, rho = std_functions.backus_average(
        vp1=vp_sst,
        vs1=vs_sst,
        rho1=rho_b_sst,
        vp2=vp_mud,
        vs2=vs_mud,
        rho2=rho_b_mud,
        f1=1.0 - shale_frac,
    )

    return vpv, vsv, vph, vsh, rho



import numpy as np
import numpy.typing as npt


def walton_smooth(
    k: npt.NDArray[np.float64],
    mu: npt.NDArray[np.float64],
    phi: npt.NDArray[np.float64],
    p_eff: npt.NDArray[np.float64],
    coord: float | npt.NDArray[np.float64] | None = None,
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    Walton Smooth Pressure-induced moduli increase at critical porosity.

    Parameters
    ----------
    k : np.ndarray
        Bulk modulus of grain mineral [Pa].
    mu : np.ndarray
        Shear modulus of grain mineral [Pa].
    phi : np.ndarray
        Critical porosity [fraction].
    p_eff : np.ndarray
        Effective Pressure = (Lithostatic - Hydrostatic) pressure [Pa].
    coord : float or np.ndarray
        Coordination number, i.e. number of grain contract per grain. If not provided a porosity based estimate is
        used [unitless].

    Returns
    -------
    tuple
        k_dry , mu_dry : np.ndarray.
        k_dry	Bulk modulus at effective pressure p,
        mu_dry	Shear modulus at effective pressure p.
    """

    n = 25.98805 * phi**2 - 43.7622 * phi + 21.6719 if coord is None else coord

    pr_min = (3 * k - 2 * mu) / (2 * (3 * k + mu))
    k_dry_num = n**2 * (1 - phi) ** 2 * mu**2 * p_eff  # Numerator in walton expression
    k_dry_denom = 18 * np.pi**2 * (1 - pr_min) ** 2  # Denominator in Walton expression
    k_dry = (k_dry_num / k_dry_denom) ** (1 / 3)  # Bulk modulus
    mu_dry = 3 / 5 * k_dry

    return k_dry, mu_dry



import numpy as np
import numpy.typing as npt

from rock_physics_open.equinor_utilities import std_functions


def min_3_voigt_reuss_hill(
    vp1: npt.NDArray[np.float64],
    vs1: npt.NDArray[np.float64],
    rhob1: npt.NDArray[np.float64],
    f1: npt.NDArray[np.float64],
    vp2: npt.NDArray[np.float64],
    vs2: npt.NDArray[np.float64],
    rhob2: npt.NDArray[np.float64],
    f2: npt.NDArray[np.float64],
    vp3: npt.NDArray[np.float64],
    vs3: npt.NDArray[np.float64],
    rhob3: npt.NDArray[np.float64],
    f3: npt.NDArray[np.float64],
) -> tuple[
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
]:
    """
    Mix of three phases by Voigt-Reuss-Hill model. The fractions should add up to 1 with input of vp, vs and rho.

    Parameters
    ----------
    vp1 : np.ndarray
        Pressure wave velocity of phase 1 [m/s].
    vs1 : np.ndarray
        Shear wave velocity of phase 1 [m/s].
    rhob1 : np.ndarray
        Bulk density of phase 1 [kg/m^3].
    f1 : np.ndarray
        Fraction of phase 1 [fraction].
    vp2 : np.ndarray
        Pressure wave velocity of phase 2 [m/s].
    vs2 : np.ndarray
        Shear wave velocity of phase 2 [m/s].
    rhob2 : np.ndarray
        Bulk density of phase 2 [kg/m^3].
    f2 : np.ndarray
        Fraction of phase 2 [fraction].
    vp3 : np.ndarray
        Pressure wave velocity of phase 3 [m/s].
    vs3 : np.ndarray
        Shear wave velocity of phase 3 [m/s].
    rhob3 : np.ndarray
        Bulk density of phase 3 [kg/m^3].
    f3 : np.ndarray
        Fraction of phase 3 [fraction].

    Returns
    -------
    tuple
        vp, vs, rhob, ai, vp_vs, k, mu : np.ndarray.
        vp: pressure wave velocity [m/s], vs: shear wave velocity [m/s], rhob: bulk density [kg/m^3],
        ai: acoustic impedance [m/s x kg/m^3], vp_vs: velocity ratio [ratio],
        k: effective bulk modulus [Pa], mu: effective shear modulus [Pa].
    """
    k1, mu1 = std_functions.moduli(vp1, vs1, rhob1)
    k2, mu2 = std_functions.moduli(vp2, vs2, rhob2)
    k3, mu3 = std_functions.moduli(vp3, vs3, rhob3)

    # Normalise the fractions to make sure they sum to one
    tot = f1 + f2 + f3
    f1 = f1 / tot
    f2 = f2 / tot
    f3 = f3 / tot

    k, mu = std_functions.multi_voigt_reuss_hill(k1, mu1, f1, k2, mu2, f2, k3, mu3, f3)

    rhob = rhob1 * f1 + rhob2 * f2 + rhob3 * f3

    vp, vs, ai, vp_vs = std_functions.velocity(k, mu, rhob)

    return vp, vs, rhob, ai, vp_vs, k, mu



from typing import Any, Protocol, TypeVar

import numpy as np
import numpy.typing as npt

_T = TypeVar("_T", bound=np.generic)

Array1D = np.ndarray[tuple[int], np.dtype[_T]]
Array2D = np.ndarray[tuple[int, int], np.dtype[_T]]
Array3D = np.ndarray[tuple[int, int, int], np.dtype[_T]]
Array4D = np.ndarray[tuple[int, int, int, int], np.dtype[_T]]


class OptCallable(Protocol):
    def __call__(
        self, x_data: npt.NDArray[np.float64], *args: Any, **kwargs: Any
    ) -> npt.NDArray[np.float64]: ...


class TMatrixCallable(Protocol):
    def __call__(
        self,
        k_min: Array1D[np.float64],
        mu_min: Array1D[np.float64],
        rho_min: Array1D[np.float64],
        k_fl: Array1D[np.float64],
        rho_fl: Array1D[np.float64],
        phi: Array1D[np.float64],
        perm: Array1D[np.float64],
        visco: Array1D[np.float64],
        alpha: Array2D[np.float64],
        v: Array2D[np.float64],
        tau: Any,
        frequency: float,
        angle: float,
        frac_inc_con: Array1D[np.float64],
        frac_inc_ani: Array1D[np.float64],
    ) -> tuple[
        Array1D[np.float64],
        Array1D[np.float64],
        Array1D[np.float64],
        Array1D[np.float64],
    ]: ...



import numpy as np

from rock_physics_open.equinor_utilities.various_utilities.types import Array3D, Array4D


def check_and_tile(
    s: Array3D[np.float64],
    t: Array4D[np.float64],
) -> tuple[Array3D[np.float64], int, int]:
    """Utility  - code that was repeated in the T-Matrix functions.

    Parameters
    ----------
    s : np.ndarray
        s parameter.
    t : np.ndarray
        t parameter.

    Returns
    -------
    tuple
        i2_i2, log_length, alpha_length : (np.ndarray, int, int).
        i2_i2: array with upper left matrices set to 1, log_length: array dimension, alpha_length: number of inclusions.
    """
    if not (
        s.ndim == 3
        and t.ndim == 4
        and np.all(np.array([s.shape[1], s.shape[2], t.shape[1]]) == t.shape[2])
        and s.shape[0] == t.shape[0]
    ):
        raise ValueError(f"{__name__}: mismatch in inputs variables dimension/shape")

    log_length = t.shape[0]
    alpha_length = t.shape[3]

    tmp = np.array(
        [
            [1, 1, 1, 0, 0, 0],
            [1, 1, 1, 0, 0, 0],
            [1, 1, 1, 0, 0, 0],
            [0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0],
        ]
    )
    i2_i2 = np.tile(tmp.reshape((1, 6, 6)), (log_length, 1, 1))

    return i2_i2, log_length, alpha_length



from pathlib import Path
from typing import cast

import numpy as np

from rock_physics_open.equinor_utilities import gen_utilities
from rock_physics_open.equinor_utilities.optimisation_utilities import (
    gen_mod_routine,
    gen_sub_routine,
    load_opt_params,
)
from rock_physics_open.equinor_utilities.various_utilities.types import Array1D

from .curvefit_t_matrix_min import curve_fit_2_inclusion_sets


def run_t_matrix_with_opt_params_petec(
    min_k: Array1D[np.float64],
    min_mu: Array1D[np.float64],
    min_rho: Array1D[np.float64],
    fl_k_orig: Array1D[np.float64],
    fl_rho_orig: Array1D[np.float64],
    fl_k_sub: Array1D[np.float64],
    fl_rho_sub: Array1D[np.float64],
    vp: Array1D[np.float64],
    vs: Array1D[np.float64],
    rhob: Array1D[np.float64],
    phi: Array1D[np.float64],
    angle: float,
    perm: float,
    visco: float,
    tau: float,
    freq: float,
    f_name: str | Path,
    fluid_sub: bool = True,
) -> tuple[
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
]:
    """
        Based on the input file with parameters for the optimally fitted model, the correct modelling version is run.
        Fluid substitution follows, in case it is selected. If not, the vp_sub and vs_sub will contain the same values as
        the input logs.
    def run_t_matrix_with_opt_params_exp(fl_k_orig, fl_rho_orig, fl_k_sub, fl_rho_sub,
                                     vp, vs, rhob,
                                     phi, vsh,
                                     angle, perm, visco, tau, freq,
                                     f_name, fluid_sub=True):
        Parameters
        ----------
        min_k : np.ndarray
            Effective mineral bulk modulus [Pa].
        min_mu : np.ndarray
            Effective mineral shear modulus [Pa].
        min_rho : np.ndarray
            Effective mineral density [kg/m^3].
        fl_k_orig : np.ndarray
            Effective in situ fluid bulk modulus [Pa].
        fl_rho_orig : np.ndarray
            Effective in situ fluid density [kg/m^3].
        fl_k_sub : np.ndarray
            Effective substituted fluid bulk modulus [Pa].
        fl_rho_sub : Effective substituted density [kg/m^3].
        vp : np.ndarray
            Compressional velocity [m/s].
        vs : np.ndarray
            Shear velocity [m/s].
        rhob : np.ndarray
            Bulk density [kg/m^3].
        phi : np.ndarray
            Porosity [fraction].
        angle : float
            Angle of symmetry plane [degrees]
        perm : float
            Permeability [mD].
        visco : float
            Viscosity [cP].
        tau : float
            Relaxation time constant [s].
        freq : float
            Signal frequency [Hz].
        f_name : str
            File name for parameter file for optimal parameters.
        fluid_sub : bool
            Boolean parameter to perform fluid substitution.

        Returns
        -------
        tuple
            Tuple of np.ndarrays: vp and vs for pressure substituted case, vp, vs and density for fluid substituted case, vp and vs for
            optimal fitted model, vp and vs residuals (observed logs minus modelled values).
    """
    _, opt_params, _ = load_opt_params(f_name)
    y_data = np.stack([vp, vs], axis=1)
    y_shape = y_data.shape
    phi, angle_, perm_, visco_, tau_, freq_, def_vpvs = cast(
        list[Array1D[np.float64]],
        gen_utilities.dim_check_vector((phi, angle, perm, visco, tau, freq, 1.0)),
    )

    rho_sub = rhob + (fl_rho_sub - fl_rho_orig) * phi
    # Set None values for inputs that will be defined in the different cases
    x_data_new = None

    opt_fcn = curve_fit_2_inclusion_sets
    # Generate x_data according to method min
    x_data = np.stack(
        (
            phi,
            min_k,
            min_mu,
            min_rho,
            fl_k_orig,
            fl_rho_orig,
            angle_,
            perm_,
            visco_,
            tau_,
            freq_,
            def_vpvs,
        ),
        axis=1,
    )
    if fluid_sub:
        x_data_new = np.stack(
            (
                phi,
                min_k,
                min_mu,
                min_rho,
                fl_k_sub,
                fl_rho_sub,
                angle_,
                perm_,
                visco_,
                tau_,
                freq_,
                def_vpvs,
            ),
            axis=1,
        )

        v_sub, v_mod, v_res = gen_sub_routine(
            opt_function=opt_fcn,
            xdata_orig=x_data,
            xdata_new=x_data_new,
            ydata=y_data,
            opt_params=opt_params,
        )
        vp_sub, vs_sub = [arr.flatten() for arr in np.split(v_sub, 2, axis=1)]
        vp_mod, vs_mod = [arr.flatten() for arr in np.split(v_mod, 2, axis=1)]
        vp_res, vs_res = [arr.flatten() for arr in np.split(v_res, 2, axis=1)]
    else:
        v_mod = gen_mod_routine(opt_fcn, x_data, y_shape, opt_params)
        vp_mod, vs_mod = [arr.flatten() for arr in np.split(v_mod, 2, axis=1)]
        vp_sub = vp
        vs_sub = vs
        vp_res = vp_mod - vp
        vs_res = vs_mod - vs

    rho_mod = min_rho * (1.0 - phi) + fl_rho_orig * phi
    rho_res = rho_mod - rhob
    ai_sub = vp_sub * rho_sub
    vpvs_sub = vp_sub / vs_sub

    return (
        vp_sub,
        vs_sub,
        rho_sub,
        ai_sub,
        vpvs_sub,
        vp_mod,
        vs_mod,
        rho_mod,
        vp_res,
        vs_res,
        rho_res,
    )



import numpy as np
from numpy.random import default_rng

from rock_physics_open.equinor_utilities.classification_functions import gen_class_stats


class TestClassStats:
    def test_class_stats(self):
        rg = default_rng(234769238476)
        obs = rg.random((100, 2))
        inp_id = rg.integers(1, 4, 100)
        class_mean, class_cov, prior_prob, class_counts, class_id = gen_class_stats(
            obs, inp_id
        )
        class_mean_ref = np.array(
            [
                [0.4450828, 0.5032985],
                [0.4959657, 0.4961214],
                [0.5020109, 0.5090125],
            ]
        )
        class_cov_ref = np.array(
            [
                [
                    [0.06409206, 0.0695167, 0.09727411],
                    [0.00732941, -0.01057039, 0.00734017],
                ],
                [
                    [0.00732941, -0.01057039, 0.00734017],
                    [0.07791386, 0.08905777, 0.07493538],
                ],
            ]
        )
        prior_prob_ref = np.array([0.28, 0.35, 0.37])
        class_count_ref = np.array([28, 35, 37])
        class_id_ref = np.array([1, 2, 3])
        np.testing.assert_almost_equal(class_mean, class_mean_ref)
        np.testing.assert_almost_equal(class_cov, class_cov_ref)
        np.testing.assert_almost_equal(prior_prob, prior_prob_ref)
        np.testing.assert_equal(class_counts, class_count_ref)
        np.testing.assert_equal(class_id, class_id_ref)



import warnings
from typing import cast

import numpy as np
import numpy.typing as npt

from rock_physics_open.equinor_utilities.gen_utilities import dim_check_vector


def gassmann(
    k_dry: npt.NDArray[np.float64],
    por: npt.NDArray[np.float64],
    k_fl: npt.NDArray[np.float64],
    k_min: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    Fluid substitution according to the Gassmann equation.

    Parameters
    ----------
    k_dry : np.ndarray
        Dry rock bulk modulus [Pa].
    por : np.ndarray
        Porosity [fraction].
    k_fl : np.ndarray
        Fluid bulk modulus.
    k_min : np.ndarray
        Mineral bulk modulus [Pa].

    Returns
    -------
    np.ndarray
        k_sat: bulk modulus for saturated rock [Pa]
    """
    k_dry, por, k_fl, k_min = cast(
        list[npt.NDArray[np.float64]],
        dim_check_vector((k_dry, por, k_fl, k_min)),
    )

    idx = np.logical_or(k_dry == k_min, por == 0)
    k_sat = np.ones(k_dry.shape) * np.nan
    b = k_dry[~idx] / (k_min[~idx] - k_dry[~idx]) + k_fl[~idx] / (
        (k_min[~idx] - k_fl[~idx]) * por[~idx]
    )
    idx1 = b < 0
    if any(idx1):
        b[idx1] = np.nan

    k_sat[~idx] = b / (1 + b) * k_min[~idx]
    k_sat[idx] = k_min[idx]

    return k_sat


def gassmann2(
    k_sat_1: npt.NDArray[np.float64],
    k_fl_1: npt.NDArray[np.float64],
    k_fl_2: npt.NDArray[np.float64],
    por: npt.NDArray[np.float64],
    k_min: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    Fluid substitution by Gassmann method with substitution of one fluid to another

    Parameters
    ----------
    k_sat_1 : np.ndarray
        bulk modulus for saturated rock with original fluid [Pa]
    k_fl_1 : np.ndarray
        bulk modulus for original fluid [Pa]
    k_fl_2 : np.ndarray
        bulk modulus for replaced fluid [Pa]
    por : np.ndarray
        porosity of rock [fraction]
    k_min : np.ndarray
        mineral bulk modulus of rock [Pa]

    Returns
    -------
    np.ndarray
        k_sat_2: bulk modulus of rock saturated with replaced fluid [Pa]
    """
    k_sat_1, k_fl_1, k_fl_2, por, k_min = cast(
        list[npt.NDArray[np.float64]],
        dim_check_vector((k_sat_1, k_fl_1, k_fl_2, por, k_min)),
    )

    idx = np.any(
        np.array([k_sat_1 == k_min, por == 0, k_fl_1 == k_min, k_fl_2 == k_min]), axis=0
    )

    k_sat_2 = np.ones(k_sat_1.shape) * np.nan

    b = (
        k_fl_2[~idx] / (por[~idx] * (k_min[~idx] - k_fl_2[~idx]))
        - k_fl_1[~idx] / (por[~idx] * (k_min[~idx] - k_fl_1[~idx]))
        + k_sat_1[~idx] / (k_min[~idx] - k_sat_1[~idx])
    )

    idx1 = b < 0
    if any(idx1):
        warn_str = (
            "{0:d} unstable solution(s) to Gassmann equation, changed to NaN".format(
                np.sum(idx1)
            )
        )
        warnings.warn(warn_str, UserWarning)
        b[idx1] = np.nan

    k_sat_2[~idx] = b / (1 + b) * k_min[~idx]
    k_sat_2[idx] = k_sat_1[idx]

    return k_sat_2


def gassmann_dry(
    k_sat: npt.NDArray[np.float64],
    por: npt.NDArray[np.float64],
    k_fl: npt.NDArray[np.float64],
    k_min: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    Dry rock properties of saturated rock by Gassmann equation

    Parameters
    ----------
    k_sat : np.ndarray
        saturated rock bulk modulus [Pa]
    por : np.ndarray
        porosity of rock [fraction]
    k_fl : np.ndarray
        bulk modulus of fluid [Pa]
    k_min : np.ndarray
        bulk modulus of mineral [Pa]

    Returns
    -------
    np.ndarray
        k_dry: dry rock bulk modulus [Pa]
    """
    k_sat, por, k_fl, k_min = cast(
        list[npt.NDArray[np.float64]],
        dim_check_vector((k_sat, por, k_fl, k_min)),
    )

    idx = np.any(np.array([k_sat == k_min, por == 0, k_fl == k_min]), axis=0)
    k_dry = np.ones(k_sat.shape)
    b = k_sat[~idx] / (k_min[~idx] - k_sat[~idx]) - k_fl[~idx] / (
        (k_min[~idx] - k_fl[~idx]) * por[~idx]
    )

    idx1 = b < 0
    if any(idx1):
        warn_str = (
            "{0:d} unstable solution(s) to Gassmann equation, changed to NaN".format(
                np.sum(idx1)
            )
        )
        warnings.warn(warn_str, UserWarning)
        b[idx1] = np.nan

    k_dry[~idx] = b / (1 + b) * k_min[~idx]
    k_dry[idx] = k_min[idx]

    return k_dry



from pathlib import Path
from shutil import copytree

import pytest

ROOT_DIR = Path(__file__).parent.resolve()
TESTDATA_DIR = ROOT_DIR.joinpath("data")


@pytest.fixture(scope="session")
def testdata() -> Path:
    return TESTDATA_DIR


@pytest.fixture(scope="session", autouse=True, name="data_dir")
def setup_rock_physics_open_test_data(
    testdata: Path, tmp_path_factory: pytest.TempPathFactory
) -> Path:
    start_dir = tmp_path_factory.mktemp("data")

    _ = copytree(testdata, start_dir, dirs_exist_ok=True)

    return start_dir



import numpy as np
import pandas as pd

from rock_physics_open.equinor_utilities.gen_utilities import dim_check_vector


class TestDimCheckVector:
    def test_dim_check_vector_ok(self):
        a = np.ones(11)
        b = 42
        c = True
        a_ref = np.ones(11)
        b_ref = np.ones_like(a) * 42
        c_ref = np.ones_like(a).astype(bool)
        a_out, b_out, c_out = dim_check_vector((a, b, c))
        np.testing.assert_equal(a_out, a_ref)
        np.testing.assert_equal(b_out, b_ref)
        np.testing.assert_equal(c_out, c_ref)

    def test_dim_check_vector_single_np_array(self):
        """Test that a single numpy array is returned as is."""
        a = np.ones(11)
        a_out = dim_check_vector(a)
        np.testing.assert_equal(a_out, a)

    def test_dim_check_vector_single_np_array_force(self):
        """Test that a single numpy array is returned as is."""
        a = np.ones(11).astype(float)
        dtype = np.dtype(int)
        a_out = dim_check_vector(a, force_type=dtype)
        np.testing.assert_equal(a_out, a)
        assert a_out.dtype == dtype

    def test_dim_check_vector_single_df(self):
        """Test that a single pandas dataframe is returned as is."""
        a = np.ones((11, 3))
        a_df = pd.DataFrame(a)
        a_out = dim_check_vector(a_df)
        pd.testing.assert_frame_equal(a_out, a_df)

    def test_dim_check_vector_single_df_force(self):
        """Test that a single pandas dataframe is returned as is."""
        a = np.ones((11, 3))
        a_df = pd.DataFrame(a).astype(float)
        dtype = np.dtype(int)
        a_out = dim_check_vector(a_df, force_type=dtype)
        pd.testing.assert_frame_equal(left=a_out, right=a_df, check_dtype=False)
        assert all(a_out.dtypes == dtype)

    def test_dim_check_vector_force(self):
        a = np.ones(11)
        b = 42
        c = True
        a_ref = np.ones(11)
        b_ref = (np.ones_like(a) * 42).astype(int)
        c_ref = np.ones_like(a).astype(bool)
        a_out, b_out, c_out = dim_check_vector((a, b, c), force_type=np.dtype(float))
        np.testing.assert_equal(a_out, a_ref)
        np.testing.assert_equal(b_out, b_ref)
        np.testing.assert_equal(c_out, c_ref)
        if isinstance(a_out, np.ndarray):
            assert a_out.dtype == a_ref.dtype
        if isinstance(b_out, np.ndarray):
            assert b_out.dtype != b_ref.dtype
        if isinstance(c_out, np.ndarray):
            assert c_out.dtype != c_ref.dtype



import numpy as np
from numpy.random import default_rng

from rock_physics_open.equinor_utilities.classification_functions import (
    gen_two_step_class_stats,
)


class TestTwoStepClass:
    def test_mahal_class_thresh(self):
        rg = default_rng(238476)
        obs = rg.random((11, 2))
        inp_class_id = np.array([3, 3, 1, 2, 3, 1, 3, 3, 2, 3, 1])
        (
            mean_class_id,
            class_cov,
            prior_prob,
            class_counts,
            class_id,
            mahal_class_id,
        ) = gen_two_step_class_stats(obs, inp_class_id, thresh=1.5)
        mahal_class_ref = np.array([3, 1, 1, 3, 0, 1, 3, 1, 2, 3, 1])
        class_id_ref = np.array([1, 2, 3])
        class_count_ref = np.array([3, 2, 5])
        prior_prob_ref = np.array([0.3, 0.2, 0.5])
        class_cov_ref = np.array(
            [
                [
                    [0.05270089, 0.03950745, 0.17263269],
                    [-0.01814353, -0.02054904, 0.11536174],
                ],
                [
                    [-0.01814353, -0.02054904, 0.11536174],
                    [0.00712315, 0.01068818, 0.08373218],
                ],
            ]
        )
        class_mean_ref = np.array(
            [
                [0.7015159, 0.8764056],
                [0.3122596, 0.2412942],
                [0.3509461, 0.5514007],
            ]
        )
        np.testing.assert_equal(mahal_class_id, mahal_class_ref)
        np.testing.assert_equal(class_id, class_id_ref)
        np.testing.assert_equal(class_counts, class_count_ref)
        np.testing.assert_almost_equal(prior_prob, prior_prob_ref)
        np.testing.assert_almost_equal(class_cov, class_cov_ref)
        np.testing.assert_almost_equal(mean_class_id, class_mean_ref)



from .opt_subst_utilities import (
    gen_mod_routine,
    gen_opt_routine,
    gen_sub_routine,
    load_opt_params,
    opt_param_info,
    opt_param_to_ascii,
    save_opt_params,
)

__all__ = [
    "gen_mod_routine",
    "gen_sub_routine",
    "gen_opt_routine",
    "opt_param_to_ascii",
    "opt_param_info",
    "save_opt_params",
    "load_opt_params",
]



from .dict_to_float import dict_value_to_float
from .dim_check_vector import dim_check_vector
from .filter_input import filter_input_log
from .filter_output import filter_output

__all__ = [
    "dict_value_to_float",
    "dim_check_vector",
    "filter_input_log",
    "filter_output",
]



import numpy as np

from rock_physics_open.equinor_utilities.various_utilities.types import (
    Array1D,
    Array2D,
    Array3D,
    Array4D,
)


def calc_t_vec(
    td: Array4D[np.float64],
    theta: Array3D[np.complex128],
    x: Array4D[np.float64],
    z: Array4D[np.complex128],
    omega: float,
    gamma: Array2D[np.float64],
    tau: Array1D[np.float64],
    k_fluid: Array1D[np.float64],
) -> Array4D[np.complex128]:
    """
    Returns the t-matrices (6x6x(numbers of connected pores)) of the
    connected pores.

    Parameters
    ----------
    td : np.ndarray
        Dry t-matrix tensors, (nx6x6x(numbers of empty cavities) matrix).
    theta : np.ndarray
        Theta-tensor (nx6x6 matrix).
    x : np.ndarray
        X-tensor (nx6x6x(numbers of empty cavities) matrix).
    z : np.ndarray
        Z-tensor (nx6x6x(numbers of empty cavities) matrix).
    omega : np.ndarray
        Frequency (2*pi*f).
    gamma : np.ndarray
        Gamma factor of all the inclusions (nx(numbers of empty cavities) vector).
    tau : np.ndarray
        Relaxation time constant (1x(numbers of empty cavities) vector).
    k_fluid : np.ndarray
        Bulk modulus of the fluid.

    Returns
    -------
    np.ndarray
        t-matrices.
    """
    if not (
        td.ndim == 4
        and x.ndim == 4
        and z.ndim == 4
        and gamma.ndim == 2
        and k_fluid.ndim == 1
        and theta.ndim == 1
        and np.all(np.array([td.shape[1:2], x.shape[1:2], z.shape[1:2]]) == td.shape[1])
        and np.all(
            np.array(
                [
                    x.shape[0],
                    z.shape[0],
                    gamma.shape[0],
                    theta.shape[0],
                    k_fluid.shape[0],
                ]
            )
            == td.shape[0]
        )
        and np.all(
            np.array([x.shape[3], z.shape[3], tau.shape[0], gamma.shape[1]])
            == td.shape[3]
        )
    ):
        raise ValueError(f"{__name__}: mismatch in inputs dimension/shape")

    log_len = k_fluid.shape[0]
    alpha_len = td.shape[3]

    # Reshape to enable broadcast
    k_fluid_ = k_fluid.reshape((log_len, 1, 1, 1))
    gamma_ = gamma.reshape((log_len, 1, 1, alpha_len))
    tau_ = tau.reshape((1, 1, 1, alpha_len))
    theta_ = theta.reshape((log_len, 1, 1, 1))

    return td + (theta_ * z + 1j * omega * tau_ * k_fluid_ * x) / (
        1 + 1j * omega * gamma_ * tau
    )



import numpy as np

from rock_physics_open.equinor_utilities.gen_utilities import filter_input_log


class TestFilterInput:
    def test_filter_input(self):
        a = np.ones(11).astype(float)
        a[4] = np.nan
        b = np.zeros_like(a)
        b[1] = 3.14
        b[9] = np.inf
        b[10] = -np.inf
        idx_ref = np.ones_like(a).astype(bool)
        idx_ref[4] = False
        idx_ref[9:] = False
        a_ref = a[idx_ref]
        b_ref = b[idx_ref]
        idx, (a_out, b_out) = filter_input_log((a, b))
        np.testing.assert_equal(idx, idx_ref)
        np.testing.assert_equal(a_out, a_ref)
        np.testing.assert_equal(b_out, b_ref)

    def test_filter_input_negative(self):
        a = np.ones(11).astype(float)
        b = np.zeros_like(a)
        b[4:6] = -1.0
        idx_ref = np.ones_like(a).astype(bool)
        idx_ref[4:6] = False
        a_ref = a[idx_ref]
        b_ref = b[idx_ref]
        idx, (a_out, b_out) = filter_input_log((a, b))
        np.testing.assert_equal(idx, idx_ref)
        np.testing.assert_equal(a_out, a_ref)
        np.testing.assert_equal(b_out, b_ref)

    def test_filter_input_no_zero(self):
        a = np.ones(11).astype(float)
        b = np.zeros_like(a)
        b[4:6] = 1.0
        idx_ref = np.zeros_like(a).astype(bool)
        idx_ref[4:6] = True
        a_ref = a[idx_ref]
        b_ref = b[idx_ref]
        idx, (a_out, b_out) = filter_input_log((a, b), no_zero=True)
        np.testing.assert_equal(idx, idx_ref)
        np.testing.assert_equal(a_out, a_ref)
        np.testing.assert_equal(b_out, b_ref)

    def test_filter_input_positive(self):
        a = np.ones(11).astype(float)
        b = np.zeros_like(a)
        b[4:6] = -1.0
        idx_ref = np.zeros_like(a).astype(bool)
        idx_ref[4:6] = True
        a_ref = a[idx_ref]
        b_ref = b[idx_ref]
        idx, (a_out, b_out) = filter_input_log((a, b), negative=True, positive=False)
        np.testing.assert_equal(idx, idx_ref)
        np.testing.assert_equal(a_out, a_ref)
        np.testing.assert_equal(b_out, b_ref)

    def test_filter_input_bool(self):
        a = np.ones(11).astype(float)
        b = np.ones_like(a).astype(bool)
        b[4:6] = False
        idx_ref = np.ones_like(a).astype(bool)
        idx_ref[4:6] = False
        a_ref = a[idx_ref]
        b_ref = b[idx_ref]
        idx, (a_out, b_out) = filter_input_log((a, b))
        np.testing.assert_equal(idx, idx_ref)
        np.testing.assert_equal(a_out, a_ref)
        np.testing.assert_equal(b_out, b_ref)

    def test_filter_input_str(self):
        a = np.ones(11).astype(float)
        a[4:6] = -1.0
        b = np.array(["NaN"] * 11)
        idx_ref = np.ones_like(a).astype(bool)
        idx_ref[4:6] = False
        a_ref = a[idx_ref]
        b_ref = b[idx_ref]
        idx, (a_out, b_out) = filter_input_log((a, b))
        np.testing.assert_equal(idx, idx_ref)
        np.testing.assert_equal(a_out, a_ref)
        np.testing.assert_equal(b_out, b_ref)



import functools
from typing import Any, Callable

import numpy as np
import numpy.typing as npt
from scipy.interpolate import RegularGridInterpolator


def generate_lookup_table(
    func: Callable[..., npt.NDArray[np.float64]],
    x: npt.NDArray[np.float64],
    y: npt.NDArray[np.float64],
    filename: str,
):
    grids2d = np.meshgrid(x, y, indexing="ij")
    grids = [g.flatten() for g in grids2d]
    z = func(*grids)
    z_grid = z.reshape(grids2d[0].shape)
    np.savez(filename, z_grid=z_grid, x=x, y=y)
    return grids2d, z_grid


@functools.lru_cache(maxsize=10)
def load_lookup_table_interpolator(
    filename: str,
) -> Callable[..., npt.NDArray[np.float64]]:
    data = np.load(filename)
    reg = RegularGridInterpolator(
        (data["x"], data["y"]), data["z_grid"], method="nearest", bounds_error=False
    )

    def _interp(
        _x: npt.NDArray[np.float64],
        _y: npt.NDArray[np.float64],
    ) -> npt.NDArray[Any]:
        _x, _y = np.asarray(_x), np.asarray(_y)
        pts = np.full((max(_x.size, _y.size), 2), fill_value=np.nan)
        pts[:, 0] = _x
        pts[:, 1] = _y
        res = reg(pts)
        if _x.ndim == 0 and _y.ndim == 0:
            return res[0]
        return res

    return _interp



from typing import final

import numpy as np
from numpy.random import default_rng

from rock_physics_open.equinor_utilities.std_functions import brie, multi_wood, wood


class TestWoodBrie:
    def setup(self):
        s_gas = np.array([1.0, 0.0, 0.0])
        s_brine = np.array([0.0, 1.0, 0.0])
        s_oil = np.array([0.0, 0.0, 1.0])
        k_gas = np.array([0.25e6, 0.25e6, 0.25e6])
        k_brine = np.array([2.8e9, 2.8e9, 2.8e9])
        k_oil = np.array([0.9e9, 0.9e9, 0.9e9])
        rho_gas = np.array([210, 210, 210])
        rho_brine = np.array([1005, 1005, 1005])

        return s_gas, s_brine, s_oil, k_gas, k_brine, k_oil, rho_gas, rho_brine

    def test_wood(self):
        s_gas, _, _, k_gas, k_brine, _, rho_gas, rho_brine = self.setup()
        k, rho = wood(s_gas, k_gas, rho_gas, k_brine, rho_brine)
        k_ref = np.array([0.25e6, 2.8e9, 2.8e9])
        rho_ref = np.array([210, 1005, 1005])
        np.testing.assert_almost_equal(k, k_ref)
        np.testing.assert_almost_equal(rho, rho_ref)

    def test_brie(self):
        s_gas, s_brine, s_oil, k_gas, k_brine, k_oil, _, _ = self.setup()
        e = 1.5
        k = brie(s_gas, k_gas, s_brine, k_brine, s_oil, k_oil, e)
        k_ref = np.array([0.25e6, 2.8e9, 0.9e9])
        np.testing.assert_almost_equal(k, k_ref)

    def test_wood_brie_mix(self):
        rg = default_rng(12345)
        s_gas = 0.5 * rg.random(11)
        s_oil = 0.5 * rg.random(11)
        s_brine = 1.0 - (s_gas + s_oil)
        k_gas = 0.25e6 * np.ones_like(s_gas)
        k_brine = 2.8e9 * np.ones_like(s_gas)
        k_oil = 0.9e9 * np.ones_like(s_gas)
        e = 1.7
        k_w = wood(s_gas, k_gas, np.ones_like(s_gas), k_brine, np.ones_like(s_gas))[0]
        k_b = brie(s_gas, k_gas, s_brine, k_brine, s_oil, k_oil, e)
        k_w_ref = np.array(
            [
                2197857.3032863,
                1577741.8284602,
                626980.6039708,
                739237.2282665,
                1277944.8143517,
                1501669.3287616,
                835514.1581449,
                2675283.4430605,
                743080.5661045,
                530843.4103878,
                2012865.2583938,
            ]
        )
        k_b_ref = np.array(
            [
                1.2081004,
                1.3510999,
                1.6787997,
                1.2305812,
                1.111577,
                1.3090595,
                1.4636728,
                1.4097729,
                1.5565275,
                1.5423466,
                2.1396782,
            ]
        )
        np.testing.assert_almost_equal(k_w, k_w_ref)
        np.testing.assert_almost_equal(k_b / 1e9, k_b_ref)


@final
class TestMultiWood:
    s_gi = 0.0
    s_oi = 0.5
    s_wi = 0.5
    oil_init_k = 492_032_461.8739312
    gas_init_k = 208_287_817.8755032
    brine_init_k = 2_814_732_817.889945
    expected_value = 837_640_292.3902907

    def test_multi_wood_array(self):
        """Test multi_wood with numpy array inputs."""
        result = multi_wood(
            fractions=[np.array(self.s_oi), np.array(self.s_gi), np.array(self.s_wi)],
            bulk_moduli=[
                np.array(self.oil_init_k),
                np.array(self.gas_init_k),
                np.array(self.brine_init_k),
            ],
        )
        expected = np.array([self.expected_value])
        np.testing.assert_almost_equal(result, expected)

    def test_multi_wood_scalar(self):
        """Test multi_wood with scalar inputs."""
        result = multi_wood(
            fractions=[self.s_oi, self.s_gi, self.s_wi],
            bulk_moduli=[self.oil_init_k, self.gas_init_k, self.brine_init_k],
        )
        expected = self.expected_value
        np.testing.assert_almost_equal(result, expected)



import warnings
from typing import Any

import numpy as np
import numpy.typing as npt
from numpy.polynomial.polynomial import polyval2d, polyval3d


def brine_properties(
    temperature: npt.NDArray[np.float64],
    pressure: npt.NDArray[np.float64],
    salinity: npt.NDArray[np.float64],
    p_nacl: npt.NDArray[np.float64] | float | None = None,  # pyright: ignore[reportUnusedParameter]
    p_kcl: npt.NDArray[np.float64] | float | None = None,  # pyright: ignore[reportUnusedParameter]
    p_cacl: npt.NDArray[np.float64] | float | None = None,  # pyright: ignore[reportUnusedParameter]
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    :param salinity: Salinity of solution as [ppm] of NaCl.
    :param pressure: Formation pressure [Pa]
    :param temperature: Temperature [°C]
    :param p_nacl: NaCl percentage, for future use
    :param p_kcl: KCl percentage, for future use
    :param p_cacl: CaCl percentage, for future use
    :return: Brine velocity vel_b [m/s], brine density den_b [kg/m^3], brine bulk modulus k_b [Pa]
    """
    vel_b = brine_primary_velocity(
        temperature=temperature,
        pressure=pressure,
        salinity=salinity,
    )
    den_b = brine_density(
        temperature=temperature,
        pressure=pressure,
        salinity=salinity,
    )
    k_b = vel_b**2 * den_b
    return vel_b, den_b, k_b


def brine_density(
    temperature: npt.NDArray[np.float64],
    pressure: npt.NDArray[np.float64],
    salinity: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    density of sodium chloride solutions, equation 27 in Batzle & Wang [1].
    :param salinity: Salinity of solution in ppm
    :param pressure: Formation pressure [Pa]
    :param temperature: Temperature [°C]
    :return: density of solution in [kg/m^3].
    """
    # Change unit of pressure to MPa
    pressure_mpa = pressure / 1.0e6
    # Change unit of salinity to fraction
    salinity_frac = salinity / 1.0e6

    coefficients = [
        [[0.668, 3e-4], [8e-5, -13e-6], [3e-6, 0.0]],
        [[0.44, -24e-4], [-33e-4, 47e-6], [0.0, 0.0]],
    ]
    water_den = water_density(temperature, pressure)
    brine_correction = (
        salinity_frac
        * polyval3d(salinity_frac, temperature, pressure_mpa, coefficients)
        * 1000.0
    )
    return water_den + brine_correction


def brine_primary_velocity(
    temperature: npt.NDArray[np.float64],
    pressure: npt.NDArray[np.float64],
    salinity: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    Primary wave velocity of sodium chloride solutions, equation 29 in Batzle & Wang [1]

    :param salinity: Salinity of solution as [ppm] of sodium chloride
    :param pressure: Formation pressure [Pa]
    :param temperature: Temperature [°C]
    :return: velocity of solution in m/s.
    """
    # Change unit for salinity from ppm to fraction
    salinity_frac = salinity / 1.0e6
    # Change the unit for pressure from Pa to MPa
    pressure_mpa = pressure / 1.0e6

    coefficients = np.zeros((3, 4, 3))
    coefficients[0, 0, 0] = 1170
    coefficients[0, 1, 0] = -9.6
    coefficients[0, 2, 0] = 0.055
    coefficients[0, 3, 0] = -8.5e-5
    coefficients[0, 0, 1] = 2.6
    coefficients[0, 1, 1] = -29e-4
    coefficients[0, 0, 2] = -0.0476
    coefficients[1, 0, 0] = 780
    coefficients[1, 0, 1] = -10
    coefficients[1, 0, 2] = 0.16
    coefficients[2, 0, 0] = -820

    return water_primary_velocity(temperature, pressure) + salinity_frac * polyval3d(
        np.sqrt(salinity_frac),
        temperature,
        pressure_mpa,
        coefficients,
    )


def water_density(
    temperature: npt.NDArray[np.float64],
    pressure: npt.NDArray[np.float64],
) -> npt.NDArray[Any]:
    """
    Density of water,, equation 27a in Batzle & Wang [1].
    :param pressure: Formation pressure [Pa]
    :param temperature: Temperature [°C]
    :return: Density of water in [kg/m^3].
    """
    # Change unit of pressure from Pa to MPa
    pressure_mpa = pressure / 1.0e6

    coefficients = [
        [1.0, 489e-6, -333e-9],
        [-8e-5, -2e-6, -2e-09],
        [-33e-7, 16e-9, 0.0],
        [1.75e-9, -13e-12, 0.0],
    ]
    return polyval2d(temperature, pressure_mpa, coefficients) * 1000.0


def water_primary_velocity(
    temperature: npt.NDArray[np.float64],
    pressure: npt.NDArray[np.float64],
) -> npt.NDArray[Any]:
    """
    Primary wave velocity of water, table 1 and equation 28 in Batzle & Wang [1].
    :param pressure: Formation pressure [Pa]
    :param temperature: Temperature [°C]
    :return: primary wave velocity of water in m/s.
    """
    # Change unit of pressure from Pa to MPa
    pressure_mpa = pressure / 1.0e6

    if np.any(pressure_mpa > 100):
        warnings.warn(
            "Calculations for water velocity is not precise for\n"
            + "pressure outside [0,100]MPa"
            + f"pressure given: {pressure}MPa",
            stacklevel=1,
        )
    coefficients = [
        [1402.85, 1.524, 3.437e-3, -1.197e-5],
        [4.871, -1.11e-2, 1.739e-4, -1.628e-6],
        [-4.783e-2, 2.747e-4, -2.135e-6, 1.237e-8],
        [1.487e-4, -6.503e-7, -1.455e-8, 1.327e-10],
        [-2.197e-7, 7.987e-10, 5.23e-11, -4.614e-13],
    ]
    return polyval2d(temperature, pressure_mpa, coefficients)


def water(
    temperature: npt.NDArray[np.float64],
    pressure: npt.NDArray[np.float64],
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    :param pressure: Formation pressure [Pa]
    :param temperature: Temperature [°C]
    :return: water_velocity [m/s], water_density [kg/m^3], water_bulk_modulus [Pa]
    """
    water_den = water_density(temperature, pressure)
    water_vel = water_primary_velocity(temperature, pressure)
    water_k = water_vel**2 * water_den
    return water_vel, water_den, water_k


def brine_viscosity(
    temperature: npt.NDArray[np.float64],
    salinity: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    Brine viscosity according to Batzle & Wang [1].

    Based on equation 32.
    """
    salinity_frac = salinity / 1.0e6
    return (
        0.1
        + 0.333 * salinity_frac
        + (1.65 + 91.9 * salinity_frac**3)
        * np.exp(-(0.42 * (salinity_frac**0.8 - 0.17) ** 2 + 0.045) * temperature**0.8)
    )



from pathlib import Path
from typing import cast

import numpy as np

from rock_physics_open.equinor_utilities import gen_utilities
from rock_physics_open.equinor_utilities.optimisation_utilities import (
    gen_mod_routine,
    load_opt_params,
)
from rock_physics_open.equinor_utilities.various_utilities.types import Array1D

from .curvefit_t_matrix_min import curve_fit_2_inclusion_sets


def run_t_matrix_forward_model_with_opt_params_petec(
    min_k: Array1D[np.float64],
    min_mu: Array1D[np.float64],
    min_rho: Array1D[np.float64],
    fl_k: Array1D[np.float64],
    fl_rho: Array1D[np.float64],
    phi: Array1D[np.float64],
    angle: float,
    perm: float,
    visco: float,
    tau: float,
    freq: float,
    f_name: str | Path,
) -> tuple[
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
]:
    """Based on the input file with parameters for the optimally fitted model, a forward modelling is done
    with inputs of mineral properties, fluid properties and porosity per sample. Other parameters (constants)
    can also be varied from their setting when the optimal parameters were found.

    Parameters
    ----------
    min_k : np.ndarray
        Effective mineral bulk modulus [Pa].
    min_mu : np.ndarray
        Effective mineral shear modulus [Pa].
    min_rho : np.ndarray
        Effective mineral density [kg/m^3].
    fl_k : np.ndarray
        Effective in situ fluid bulk modulus [Pa].
    fl_rho : np.ndarray
        Effective in situ fluid density [kg/m^3].
    phi : np.ndarray
        Porosity [fraction].
    angle : float
        Angle of symmetry plane [degrees]
    perm : float
        Permeability [mD].
    visco : float
        Viscosity [cP].
    tau : float
        Relaxation time constant [s].
    freq : float
        Signal frequency [Hz].
    f_name : str
        File name for parameter file for optimal parameters.

    Returns
    -------
    tuple
        Tuple of np.ndarrays: vp [m/s], vs [m/s], rho [kg/m^3], ai [kg/m^3 x m/s], vp/vs [fraction] for forward model.
    """
    opt_type, opt_params, _ = load_opt_params(f_name)
    phi, angle_, perm_, visco_, tau_, freq_, def_vpvs = cast(
        list[Array1D[np.float64]],
        gen_utilities.dim_check_vector((phi, angle, perm, visco, tau, freq, 1.0)),
    )
    rho_mod = min_rho * (1.0 - phi) + fl_rho * phi
    y_shape = (phi.shape[0], 2)

    if opt_type != "min":
        raise TypeError(
            f"{__file__}: incorrect type of optimal parameter input file, must come from PETEC optimisation"
        )
    # No Need for preprocessing
    opt_fcn = curve_fit_2_inclusion_sets
    # Generate x_data according to method min
    x_data = np.stack(
        (
            phi,
            min_k,
            min_mu,
            min_rho,
            fl_k,
            fl_rho,
            angle_,
            perm_,
            visco_,
            tau_,
            freq_,
            def_vpvs,
        ),
        axis=1,
    )
    v_mod = gen_mod_routine(opt_fcn, x_data, y_shape, opt_params)
    if not isinstance(v_mod, np.ndarray):  # pyright: ignore[reportUnnecessaryIsInstance] | kept for backwards compatibility
        raise ValueError(f"{__file__}: no solution to forward model")
    vp_mod, vs_mod = [arr.flatten() for arr in np.split(v_mod, 2, axis=1)]
    vpvs_mod = vp_mod / vs_mod
    ai_mod = vp_mod * rho_mod
    return vp_mod, vs_mod, rho_mod, ai_mod, vpvs_mod



import numpy as np
import pytest

from rock_physics_open.equinor_utilities.std_functions import rho_b, rho_m


class TestRho:
    def test_rhob(self):
        phi = np.linspace(0.0, 0.25, 10)
        rho_fluid = 1000 * np.ones_like(phi)
        rho_matrix = 3000 * np.ones_like(phi)

        rhob = rho_b(phi, rho_fluid, rho_matrix)
        np.testing.assert_almost_equal(rhob, np.linspace(3000, 2500, 10))

    def test_rhom(self):
        phi = np.linspace(0.0, 0.25, 10)
        rho_cem = 2000 * np.ones_like(phi)
        rho_min = 3000 * np.ones_like(phi)
        frac_cem = 0.05 * np.ones_like(phi)

        rho_mat = rho_m(frac_cem, phi, rho_cem, rho_min)
        rho_expected = np.array(
            [
                2950.0,
                2948.5714286,
                2947.0588235,
                2945.4545455,
                2943.75,
                2941.9354839,
                2940.0,
                2937.9310345,
                2935.7142857,
                2933.3333333,
            ]
        )
        np.testing.assert_almost_equal(rho_mat, rho_expected)

    def test_phi_one(self):
        phi = np.array([1.0])
        rho_cem = 2000 * np.ones_like(phi)
        rho_min = 3000 * np.ones_like(phi)
        frac_cem = 0.05 * np.ones_like(phi)

        with pytest.warns(UserWarning, match="phi out of range in 1 sample"):
            r_m = rho_m(frac_cem, phi, rho_cem, rho_min)
        assert np.isnan(r_m[0])



import os
import sys
from collections.abc import Sequence
from typing import Any, Literal, cast

import numpy as np
import numpy.typing as npt


def disp_result_stats(
    title: str, arr: Sequence[Any], names_arr: list[str], **kwargs: Any
) -> None:
    """
    Display results utilizing tkinter.

    Parameters
    ----------
    title : Title.
    arr : items to display
    names_arr : array of names.
    """
    from tkinter import END, Entry, PhotoImage, Tk

    class Table:
        def __init__(
            self, tk_root: Tk, no_rows: int, no_cols: int, info: npt.NDArray[Any]
        ):
            # code for creating table
            str_len = np.vectorize(len)
            text_justify: list[Literal["center", "left"]] = ["center", "left"]
            text_weight: list[str] = ["bold", "normal"]
            for i in range(no_rows):
                weigh = text_weight[cast(int, np.sign(i))]
                for j in range(no_cols):
                    just = text_justify[cast(int, np.sign(i))]
                    max_len = np.max(str_len(info[:, j]))
                    self.e: Entry = Entry(
                        tk_root,
                        width=max_len + 2,
                        fg="black",
                        font=("Consolas", 12, weigh),
                        justify=just,
                    )
                    self.e.grid(row=i, column=j)
                    self.e.insert(END, info[i][j])

    values_only = kwargs.pop("values_only", False)
    root = Tk(**kwargs)
    root.title(title)
    if values_only:
        info_array = np.zeros((len(arr) + 1, 2)).astype(str)
        info_array[0, :] = ["Property", "Value"]
        for k in range(len(arr)):
            info_array[k + 1, 0] = f"{names_arr[k]}"
            info_array[k + 1, 1] = f"{arr[k]:.3g}"
    else:
        info_array = np.zeros((len(arr) + 1, 6)).astype(str)
        info_array[0, :] = ["Var", "Min", "Mean", "Max", "No. NaN", "No. of Inf"]
        for k in range(len(arr)):
            info_array[k + 1, 0] = f"{names_arr[k]}"
            info_array[k + 1, 1] = f"{np.nanmin(np.asarray(arr[k])):.3g}"
            info_array[k + 1, 2] = f"{np.nanmean(np.asarray(arr[k])):.3g}"
            info_array[k + 1, 3] = f"{np.nanmax(np.asarray(arr[k])):.3g}"
            info_array[k + 1, 4] = f"{np.sum(np.isnan(np.asarray(arr[k])))}"
            info_array[k + 1, 5] = f"{np.sum(np.isinf(np.asarray(arr[k])))}"

    _ = Table(root, info_array.shape[0], info_array.shape[1], info_array)

    root.update_idletasks()
    window_height = root.winfo_height()
    window_width = root.winfo_width()
    screen_width = root.winfo_screenwidth()
    screen_height = root.winfo_screenheight()

    x_coordinate = int((screen_width / 2) - (window_width / 2))
    y_coordinate = int((screen_height / 2) - (window_height / 2))

    root.geometry(
        "{}x{}+{}+{}".format(window_width, window_height, x_coordinate, y_coordinate)
    )

    if sys.platform.startswith("win"):
        root.iconbitmap(os.path.join(os.path.dirname(__file__), "Equinor_logo.ico"))
    else:
        logo = PhotoImage(
            file=os.path.join(os.path.dirname(__file__), "Equinor_logo.gif")
        )
        root.wm_iconphoto(True, logo)

    root.mainloop()



import numpy as np
import numpy.typing as npt

NULL_CLASS = 0


def gen_class_stats(
    obs: npt.NDArray[np.float64],
    class_val: npt.NDArray[np.int64],
) -> tuple[
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.int64],
    npt.NDArray[np.int64],
]:
    """
    Generate statistics - mean, covariance and prior probability - for each
    class in the training data.	The observations are an n x m array, where n
    is the number of observations and m is the number of variables. With p
    classes the returned mean value will be an array of dimension p x m,
    covariance m x m x p and the class_id and prior probability p length vector.
    class_mean, class_cov, prior_prob, class_counts, class_id = gen_class_stats(obs, class_val).

    Parameters
    ----------
    obs : np.ndarray
        An nxm array of data samples (observations).
    class_val : np.ndarray
        n length vector with class ID of the observations. Assumed to
        be integer.

    Returns
    -------
    tuple
        class_mean, class_cov, prior_prob, class_counts, class_id : (np.ndarray, np.ndarray, np.ndarray, np.ndarray,
        np.ndarray).
    """

    n, m = obs.shape
    # Find number of classes. If class_val input is not integer, raise an exception
    if not (
        isinstance(class_val, np.ndarray)  # pyright: ignore[reportUnnecessaryIsInstance] | Kept for backward compatibility
        and issubclass(class_val.dtype.type, np.integer)  # pyright: ignore[reportUnnecessaryIsInstance] | Kept for backward compatibility
    ):
        raise ValueError(f"{__file__}: class values are not discrete numbers")

    class_id, class_counts = np.unique(class_val, return_counts=True)
    # Remove Null class
    idx_null = np.where(class_id == NULL_CLASS)
    class_id = np.delete(class_id, idx_null)
    class_counts = np.delete(class_counts, idx_null)
    p = class_id.shape[0]

    # Very simple prior probability - number of observations in each class
    # divided by total number of observations
    prior_prob = class_counts / n

    # Assign output arrays
    class_mean = np.zeros((p, m))
    class_cov = np.zeros((m, m, p))

    for i in range(len(class_id)):
        idx = class_val == class_id[i]
        class_mean[i, :] = np.mean(obs[idx, :], axis=0)
        class_cov[:, :, i] = np.cov(obs[idx, :], rowvar=False)

    return class_mean, class_cov, prior_prob, class_counts, class_id



from rock_physics_open.equinor_utilities.gen_utilities import dict_value_to_float


class TestDictToFloat:
    def test_dict_to_float(self):
        inp_dict = {"a": "1.0", "b": "2", "d": "[1.2, 4.5]", "e": "-3"}
        ref_dict = {"a": 1.0, "b": 2.0, "d": [1.2, 4.5], "e": -3.0}
        out_dict = dict_value_to_float(inp_dict)
        assert out_dict == ref_dict



from typing import TypeVar

import numpy as np

from rock_physics_open.equinor_utilities.various_utilities.types import Array3D, Array4D

_T = TypeVar(
    "_T",
    Array3D[np.float64],
    Array4D[np.float64],
)


def iso_av_vec(t: _T) -> _T:
    """Returns a (nx6x6) matrix t_bar averaged over all the orientations (isotropy).


    t_bar = np.array([

        [c11, c12, c12, 0, 0, 0],

        [c12, c11, c12, 0, 0, 0],

        [c12, c12, c11, 0, 0, 0],

        [0, 0, 0, 2 * c44, 0, 0],

        [0, 0, 0, 0, 2 * c44, 0],

        [0, 0, 0, 0, 0, 2 * c44]

    ])


    Parameters
    ----------
    t : np.ndarray
        An nx6x6 matrix which has a HTI symmetry.


    Returns
    -------
    np.ndarray
        Averaged value.
    """
    t_bar = np.zeros_like(t)

    lambda_var = (
        t[:, 0, 0] + t[:, 2, 2] + 5 * t[:, 0, 1] + 8 * t[:, 0, 2] - 2 * t[:, 3, 3]
    ) / 15
    mu = (
        7 * t[:, 0, 0]
        + 2 * t[:, 2, 2]
        - 5 * t[:, 0, 1]
        - 4 * t[:, 0, 2]
        + 6 * t[:, 3, 3]
    ) / 30
    c11 = lambda_var + 2 * mu
    c12 = lambda_var
    c44 = mu

    for i in range(3):
        t_bar[:, i, i] = c11
        t_bar[:, i + 3, i + 3] = 2 * c44
    for i in range(2):
        t_bar[:, 0, i + 1] = c12
        t_bar[:, 1, 2 * i] = c12
        t_bar[:, 2, i] = c12

    return t_bar



from typing import Literal

import numpy as np

from rock_physics_open.equinor_utilities.various_utilities.types import Array4D

from .iso_av import iso_av_vec


def iso_av_all_vec(
    x: Array4D[np.float64],
    case_iso: Literal[0, 1, 2],
) -> Array4D[np.float64]:
    """Returns an multi dimensional matrix with isotropic elements. (nx6x6x(numbers of inclusions) matrix).
    Not used in the present implementation - direct call to iso_av_vec from main function instead


    Parameters
    ----------
    x : np.ndarray
        An nx6x6x(numbers of inclusions) matrix.
    case_iso : int
        Control parameter.

    Returns
    -------
    np.ndarray,
        Isotropic values.

    Examples
    --------
    caseIso : control parameter
    if caseIso = 0 then all the pore types are isotropic,
    if caseIso = 1 then there is a mix of isotropic and anisotropic pores,
    if caseIso = 2 then all the pore types are anisotropic.

    Notes
    -----
    HFLE 01.11.2020: Anisotropic inclusions are set to have the same aspect ratio as the isotropic ones.
    """
    if not (x.ndim == 4 and x.shape[1] == x.shape[2]):
        raise ValueError(f"{__name__}: mismatch in inputs variables dimension/shape")

    no_inclusions = x.shape[3]
    x_iso = np.zeros(x.shape)

    # if caseIso = 0 then all the pore types are isotropic
    if case_iso == 0:
        for j in range(no_inclusions):
            x_iso[:, :, j] = iso_av_vec(x[:, :, j])
    # if caseIso = 1 then all the pore types are isotropic
    elif case_iso == 1:
        for j in range(no_inclusions):
            x_iso[:, :, j] = iso_av_vec(x[:, :, j])

        x_iso[:, :, (no_inclusions // 2) : no_inclusions] = x[
            :, :, (no_inclusions // 2) : no_inclusions
        ]
    elif case_iso == 2:
        x_iso = x

    return x_iso



import numpy as np
import numpy.typing as npt


def backus_average(
    vp1: npt.NDArray[np.float64],
    vs1: npt.NDArray[np.float64],
    rho1: npt.NDArray[np.float64],
    vp2: npt.NDArray[np.float64],
    vs2: npt.NDArray[np.float64],
    rho2: npt.NDArray[np.float64],
    f1: npt.NDArray[np.float64],
) -> tuple[
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
]:
    """
    Backus average for a combination of two phases. The individual phases are isotropic
    but the resulting effective medium is not.

    Parameters
    ----------
    vp1 : np.ndarray
        Pressure wave velocity for phase 1 [m/s].
    vs1 : np.ndarray
        Shear wave velocity for phase 1 [m/s].
    rho1 : np.ndarray
        Density for phase 1 [kg/m^3].
    vp2 : np.ndarray
        Pressure wave velocity for phase 2 [m/s].
    vs2 : np.ndarray
        Shear wave velocity for phase 2 [m/s].
    rho2 : np.ndarray
        Density for phase 2 [kg/m^3].
    f1 : np.ndarray
        Fraction of phase 1.

    Returns
    -------
    tuple
        vpv, vsv, vph, vsh, rho : np.ndarray
        vpv: vertical pressure velocity, vsv: vertical shear velocity, vph: horizontal pressure velocity,
        vsh: horizontal shear velocity, rho: density.
    """

    a = (
        4 * f1 * rho1 * vs1**2 * (1 - vs1**2 / vp1**2)
        + 4 * (1 - f1) * rho2 * vs2**2 * (1 - (vs2 / vp2) ** 2)
        + (f1 * (1 - 2 * (vs1 / vp1) ** 2) + (1 - f1) * (1 - 2 * (vs2 / vp2) ** 2)) ** 2
        * (1 / (f1 / (rho1 * vp1**2) + (1 - f1) / (rho2 * vp2**2)))
    )

    c = 1 / (f1 / (rho1 * vp1**2) + (1 - f1) / (rho2 * vp2**2))
    d = 1 / (f1 / (rho1 * vs1**2) + (1 - f1) / (rho2 * vs2**2))

    m = f1 * rho1 * vs1**2 + (1 - f1) * rho2 * vs2**2

    rho = f1 * rho1 + (1 - f1) * rho2

    vpv = np.sqrt(c / rho)
    vsv = np.sqrt(d / rho)
    vph = np.sqrt(a / rho)
    vsh = np.sqrt(m / rho)

    return vpv, vsv, vph, vsh, rho
